"""Base forest-related models using LightGBM as the gradient booster."""
from abc import abstractmethod
import os
from typing import Any, ClassVar, Dict, Optional, Sequence, Tuple, Union, cast

import lightgbm as lgb
from marshmallow import fields
import numpy as np

from bitfount.data.databunch import BitfountDataBunch
from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasources.base_source import BaseSource
from bitfount.models.base_models import ClassifierMixIn, _BaseModel
from bitfount.types import T_FIELDS_DICT

# TODO: [BIT-1041] Introduce non-implementation base model
from bitfount.utils import delegates


@delegates()
class BaseLGBMRandomForest(_BaseModel):
    """Implements an (optionally Gradient Boosted) Random Forest from LightGBM."""

    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "gradient_boosting": fields.Boolean(),
        "num_leaves": fields.Integer(allow_none=True),
        "max_depth": fields.Integer(allow_none=True),
        "subsample_for_bin": fields.Integer(allow_none=True),
        "num_iterations": fields.Integer(allow_none=True),
        "learning_rate": fields.Float(allow_none=True),
        "reg_alpha": fields.Float(allow_none=True),
        "reg_lambda": fields.Float(allow_none=True),
        "bagging_freq": fields.Integer(allow_none=True),
        "bagging_fraction": fields.Float(allow_none=True),
        "feature_fraction": fields.Float(allow_none=True),
        "early_stopping_rounds": fields.Integer(allow_none=True),
        "verbose": fields.Integer(allow_none=True),
        "min_split_gain": fields.Float(allow_none=True),
    }

    def __init__(
        self,
        gradient_boosting: bool = False,
        num_leaves: Optional[int] = None,
        max_depth: Optional[int] = None,
        subsample_for_bin: Optional[int] = None,
        num_iterations: Optional[int] = None,
        learning_rate: Optional[float] = None,
        reg_alpha: Optional[float] = None,
        reg_lambda: Optional[float] = None,
        bagging_freq: Optional[float] = None,
        bagging_fraction: Optional[float] = None,
        feature_fraction: Optional[float] = None,
        early_stopping_rounds: Optional[int] = None,
        verbose: Optional[int] = None,
        min_split_gain: Optional[float] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        self.model_type = "gbdt" if gradient_boosting is True else "rf"
        self.num_iterations = 10000 if num_iterations is None else num_iterations
        self.learning_rate = 0.05 if learning_rate is None else learning_rate
        self.num_leaves = 50 if num_leaves is None else num_leaves
        self.max_depth = 10 if max_depth is None else max_depth
        self.subsample_for_bin = (
            240000 if subsample_for_bin is None else subsample_for_bin
        )
        self.reg_alpha = 0.5 if reg_alpha is None else reg_alpha
        self.reg_lambda = 0.5 if reg_lambda is None else reg_lambda
        self.bagging_freq = 1 if bagging_freq is None else bagging_freq
        self.bagging_fraction = 0.5 if bagging_fraction is None else bagging_fraction
        self.feature_fraction = 0.5 if feature_fraction is None else feature_fraction
        self.early_stopping_rounds = (
            100 if early_stopping_rounds is None else early_stopping_rounds
        )
        self.verbose = 100 if verbose is None else verbose
        self.min_split_gain = 0.025 if min_split_gain is None else min_split_gain
        self.test_set = None
        self.objective: Optional[str] = None
        self.training_metrics: Optional[Sequence[str]] = None

    @abstractmethod
    def _set_objective(self) -> None:
        """Set training objective."""
        raise NotImplementedError

    @abstractmethod
    def _set_training_metrics(self) -> None:
        """Set training metrics."""
        raise NotImplementedError

    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Serialize model."""
        self._model.save_model(filename, num_iteration=self._model.best_iteration)

    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Deserialize model."""
        self._model = lgb.Booster(model_file=filename)

    def evaluate(
        self, test_dl: Optional[BitfountDataLoader] = None, *args: Any, **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Perform inference on test set and save dictionary of metrics."""
        if test_dl is None:
            if isinstance(self.test_dl, BitfountDataLoader):
                test_dl = self.test_dl
            else:
                raise ValueError("There is no test data to evaluate the model on.")

        test_df = test_dl.get_x_dataframe()
        if isinstance(test_df, tuple):
            raise ValueError(
                "Multiple dataframes retrieved unexpectedly; "
                "this model does not support combination tabular and image data."
            )

        test_preds = np.zeros(test_df.shape[0])
        best_iteration = self._model.best_iteration
        test_preds += self._model.predict(test_df, num_iteration=best_iteration)
        test_target = test_dl.get_y_dataframe().to_numpy()

        return test_preds, test_target

    def get_params(self) -> Dict:
        """Create an instance of the model."""
        params = {
            "objective": self.objective,
            "boosting_type": self.model_type,
            "learning_rate": self.learning_rate,
            "num_leaves": self.num_leaves,
            "max_depth": self.max_depth,
            "subsample_for_bin": self.subsample_for_bin,
            "reg_alpha": self.reg_alpha,
            "reg_lambda": self.reg_lambda,
            "min_split_gain": self.min_split_gain,
            "bagging_fraction": self.bagging_fraction,
            "bagging_freq": self.bagging_freq,
            "feature_fraction": self.feature_fraction,
            "random_state": self.seed,
            "metric": self.training_metrics,
            "verbose": self.verbose,
        }

        return params

    def _create_dataset(self) -> Tuple[lgb.Dataset, Optional[lgb.Dataset]]:
        """Create LightGBM Dataset object for better memory management."""
        x_train = self.train_dl.get_x_dataframe()
        if isinstance(x_train, tuple):
            raise ValueError(
                "Multiple dataframes retrieved unexpectedly; "
                "this model does not support combination tabular and image data."
            )

        y_train = self.train_dl.get_y_dataframe().astype("int64")
        train_df = lgb.Dataset(
            data=x_train,
            label=y_train,
            free_raw_data=True,
        )

        weights_col = self.datastructure.loss_weights_col
        if weights_col is not None:
            train_df = train_df.set_weight(weight=list(x_train[weights_col]))
        if isinstance(self.validation_dl, BitfountDataLoader):
            x_valid = self.validation_dl.get_x_dataframe()
            if isinstance(x_valid, tuple):
                raise ValueError(
                    "Multiple dataframes retrieved unexpectedly; "
                    "this model does not support combination tabular and image data."
                )

            y_valid = self.validation_dl.get_y_dataframe().astype("int64")
            # For LightGBM DataFrame.dtypes for label must be int, float or bool

            validation_df = lgb.Dataset(
                data=x_valid,
                label=y_valid,
                free_raw_data=True,
                reference=train_df,
            )
            if weights_col is not None:
                validation_df = validation_df.set_weight(
                    weight=list(x_valid[weights_col])
                )
        else:
            validation_df = None

        return train_df, validation_df

    def fit(self, data: Optional[BaseSource] = None, *args: Any, **kwargs: Any) -> None:
        """Trains a model using the training set provided by the BaseSource object."""
        if data:
            if self.datastructure.query:
                table_schema = self.datastructure._override_schema()
                self.databunch = BitfountDataBunch(
                    data_structure=self.datastructure,
                    schema=table_schema,
                    datasource=data,
                )
            elif self.datastructure.table:
                table_name = self.datastructure.get_table_name()
                table_schema = self.schema.get_table_schema(table_name)
                self._add_datasource_to_schema(data)

            if self.objective == "binary":
                # The casts here are to assuage mypy because it (incorrectly) asserts
                # that a subclass of both ClassifierMixIn and LightGBMModel can't exist.
                # We utilise a subclass of both in the tests to assure ourselves.
                if isinstance(cast(ClassifierMixIn, self), ClassifierMixIn):
                    cast(ClassifierMixIn, self).set_number_of_classes(table_schema)
                else:
                    raise TypeError(
                        "Training objective is classification but this model does not "
                        "inherit from ClassifierMixIn"
                    )
            self._set_dataloaders()

        else:
            raise ValueError(
                "No data provided. This model can only be trained on local data."
            )

        params = self.get_params()
        lgb_train, lgb_valid = self._create_dataset()
        if lgb_valid:
            valid_sets = [lgb_valid, lgb_train]
            valid_names = ["valid", "train"]
        else:
            valid_sets = [lgb_train]
            valid_names = ["train"]

        model = lgb.train(
            params=params,
            num_boost_round=self.num_iterations,
            train_set=lgb_train,
            valid_sets=valid_sets,
            valid_names=valid_names,
            verbose_eval=self.verbose,
            early_stopping_rounds=self.early_stopping_rounds,
            init_model=self._model,
        )
        self._model: lgb.Booster = model



"""Tests config.py."""
import inspect

from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
import pytest
from pytest import fixture

from bitfount.federated.transport.config import MessageServiceConfig
from bitfount.federated.transport.protos.messages_pb2_grpc import MessageServiceStub
from tests.utils.helper import unit_test


@unit_test
class TestMessageServiceConfig:
    """Tests for MessageServiceConfig."""

    @fixture
    def test_url(self) -> str:
        """URL for tests."""
        return "not.a.real.url.com"

    def test_ms_config_raises_value_error_if_production_url_and_tls_disabled(
        self, monkeypatch: MonkeyPatch
    ) -> None:
        """Ensure TLS can't be disabled with production hub url."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "production")
        with pytest.raises(ValueError):
            MessageServiceConfig(tls=False)

    def test_ms_config_raises_value_error_if_staging_url_and_tls_disabled(
        self, monkeypatch: MonkeyPatch
    ) -> None:
        """Ensure TLS can't be disabled with staging hub url."""
        monkeypatch.setenv("BITFOUNT_ENVIRONMENT", "staging")
        with pytest.raises(ValueError):
            MessageServiceConfig(tls=False)

    def test_ms_config_issues_warning_if_tls_disabled_with_custom_url(
        self, caplog: LogCaptureFixture
    ) -> None:
        """Ensure that warning is issued if tls is disabled with non-bitfount URL.

        NB: There should only be one record in caplog.records.
        """
        MessageServiceConfig(url="custom.ms.url.com")
        for record in caplog.records:
            assert record.levelname == "WARNING"
            assert record.getMessage() == "Message service communication without TLS."

    async def test_creates_message_service_stub_secure(self, test_url: str) -> None:
        """Tests create_message_service_stub with TLS."""
        ms_config = MessageServiceConfig(tls=False, url=test_url)
        stub = await ms_config.stub
        assert isinstance(stub, MessageServiceStub)

    async def test_creates_message_service_stub_insecure(self, test_url: str) -> None:
        """Tests create_message_service_stub without TLS."""
        ms_config = MessageServiceConfig(tls=True, url=test_url)
        stub = await ms_config.stub
        assert isinstance(stub, MessageServiceStub)

    def test_stub_property_is_async(self, test_url: str) -> None:
        """Tests that the MessageServiceConfig.stub property is async."""
        # Check is property (properties exist on the class)
        assert isinstance(MessageServiceConfig.stub, property)

        # Check property getter is async
        assert inspect.iscoroutinefunction(MessageServiceConfig.stub.fget)



"""Utility functions concerning data."""
from dataclasses import dataclass
import datetime
import hashlib
import json
import logging
from typing import Any, List, Mapping, Optional, Union

import pandas as pd
from pandas.core.dtypes.common import pandas_dtype
import sqlalchemy
from sqlalchemy import create_engine, inspect

from bitfount.data.exceptions import (
    DatabaseMissingTableError,
    DatabaseSchemaNotFoundError,
    DatabaseUnsupportedQueryError,
)
from bitfount.types import _DtypesValues

logger = logging.getLogger(__name__)


@dataclass
class DatabaseConnection:
    """Encapsulates database connection information for a `BaseSource`.

    If a `query` is provided or if `table_name` only has one table, the database will be
    queried for the data, after which the database connection will be closed and the
    resulting DataFrame will be used and stored in the `BaseSource`.

    :::danger

    If you are creating a multi-table Pod, ensure that the connection you provide only
    has access to the schemas and tables you wish to share and that this access has
    suitably restricted permissions i.e. `SELECT` only.

    `table_names` limits the Pod schema to only those tables you specify but it does
    not prevent a Modeller from accessing other tables in the schema or indeed other
    tables in other schemas by guessing their names.

    If only a single table is provided or a query is provided to combine multiple tables
    into one table, the Modeller will have no access to the database.

    :::

    Args:
        con: A database URI provided as a string or a SQLAlchemy Engine. This should
            include the database name, user, password, host, port, etc.
        db_schema: The database schema to use. If not provided, the default schema will
            be used.
        query: The SQL query to be executed as a string.
        table_names: Name(s) of SQL table(s) in database.

    Attributes:
        multi_table: Whether or not the database connection is for multiple tables.

    Raises:
        DatabaseMissingTableError: If `schema` (or the default schema if not provided)
            does not contain any tables or any of the specified tables can't be found in
            the schema.
        DatabaseSchemaNotFoundError: If `schema` is provided but can't be found in the
            database.
        DatabaseModificationError: If `query` is provided and contains an 'INTO' clause.
        ValueError: If both `query` and `table_names` are provided.
    """

    def __init__(
        self,
        con: Union[str, sqlalchemy.engine.base.Engine],
        db_schema: Optional[str] = None,
        query: Optional[str] = None,
        table_names: Optional[List[str]] = None,
    ) -> None:

        # Create SQLAlchemy engine from database URI
        if isinstance(con, str):
            self.con: sqlalchemy.engine.base.Engine = create_engine(con)
        else:
            self.con = con

        self.query = query
        self.table_names = table_names if table_names else []

        inspector = inspect(self.con)

        # Ensure the provided schema exists, if not provided use the default schema
        if db_schema:
            schemas = inspector.get_schema_names()
            if db_schema not in schemas:
                raise DatabaseSchemaNotFoundError(
                    f"Schema '{db_schema}' not found in database."
                )
        self.db_schema: str = db_schema if db_schema else inspector.default_schema_name

        # Default is that we are dealing with a single table
        self.multi_table: bool = False

        # Raise an error if both query and table_names are provided
        if self.query and self.table_names:
            raise ValueError("Cannot specify both query and table_names.")

        # If table_names is provided, ensure that the schema contains all of the
        # specified tables
        elif self.table_names:
            db_table_names = inspector.get_table_names(schema=self.db_schema)
            for table in self.table_names:
                if table not in db_table_names:
                    raise DatabaseMissingTableError(
                        f"Table '{table}' not found in schema '{self.db_schema}'."
                    )

            if len(self.table_names) == 1:
                logger.info(
                    f"Single table {self.table_names[0]} provided. Reading data."
                )
            else:
                self.multi_table = True
                logger.info("Restricting schema to only specified tables.")
                logger.warning("Modeller needs to provide query in DataStructure.")

        # Nothing needs to be done here if a query was provided
        elif self.query:
            if "into" in self.query.lower():
                # 'SELECT * INTO' or 'INSERT INTO' would create a new table which is
                # not allowed
                raise DatabaseUnsupportedQueryError(
                    "Invalid query provided. "
                    "'SELECT * INTO' and 'INSERT INTO' are unsupported statements."
                )
            else:
                logger.info("Query provided. Reading data.")
        # If no query or table_names were provided, read all the tables from the schema
        else:
            logger.info(
                "No query or table names provided. "
                "Using all tables in the database for the schema. "
            )
            db_table_names = inspector.get_table_names(schema=self.db_schema)
            if db_table_names:
                self.table_names = db_table_names
            else:
                raise DatabaseMissingTableError(
                    f"No tables found in schema '{self.db_schema}'."
                )
            if len(self.table_names) > 1:
                self.multi_table = True
            logger.warning("Modeller needs to provide query in DataStructure.")


def _generate_dtypes_hash(dtypes: Mapping[str, Any]) -> str:
    """Generates a hash of a column name -> column type mapping.

    Uses column names and column dtypes to generate the hash. DataFrame contents
    is NOT used.

    SHA256 is used for hash generation.

    Args:
        dtypes: The mapping to hash.

    Returns:
        The hexdigest of the mapping hash.
    """
    dtypes = {k: str(v) for k, v in dtypes.items()}
    str_rep: str = json.dumps(dtypes, sort_keys=True)
    return _hash_str(str_rep)


def _hash_str(to_hash: str) -> str:
    """Generates a sha256 hash of a given string.

    Uses UTF-8 to encode the string before hashing.
    """
    return hashlib.sha256(to_hash.encode("utf-8")).hexdigest()


def _convert_python_dtypes_to_pandas_dtypes(
    dtype: _DtypesValues, col_name: str
) -> _DtypesValues:
    """Convert the python dtypes to pandas dtypes.

    Used for the schema generation for the `DatabaseSource`.
    """
    if dtype == str:
        return pd.StringDtype()
    # SQLAlchemy is able to infer the dtype of a column from the data in the column
    # and exposes this using the `python_type` attribute which includes types in the
    # python `datetime` standard library.
    elif dtype == datetime.date or dtype == datetime.datetime:
        return pd.StringDtype()
    else:
        try:
            return pandas_dtype(dtype)
        except Exception:
            raise ValueError(f"Data type {dtype} not recognised for column {col_name}")



"""Pytest fixtures for data tests."""
from pathlib import Path

import pandas as pd
from pytest import fixture
import sqlalchemy

from bitfount.data.datasets import _BitfountDataset, _IterableBitfountDataset
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import DataSplit, SemanticType
from bitfount.data.utils import DatabaseConnection
from tests.utils.helper import TABLE_NAME, create_dataset, create_segmentation_dataset


@fixture
def dataframe() -> pd.DataFrame:
    """Underlying dataframe for single image datasets."""
    return create_dataset(image=True)


@fixture
def tabular_dataset(dataframe: pd.DataFrame) -> _BitfountDataset:
    """Basic tabular dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe, ignore_cols=["image"])
    datasource.load_data()

    schema = BitfountSchema()
    schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target, ignore_cols=["image"], table=TABLE_NAME
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _BitfountDataset(
        datasource=datasource,
        target=target,
        selected_cols=datastructure.selected_cols,
        data_split=DataSplit.TRAIN,
        schema=schema.get_table_schema(TABLE_NAME),
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
    )


@fixture
def db_dataset(db_session: sqlalchemy.engine.base.Engine) -> _BitfountDataset:
    """Basic tabular dataset for tests as fixture."""
    target = "TARGET"
    db_conn = DatabaseConnection(db_session, table_names=["dummy_data", "dummy_data_2"])
    datasource = DatabaseSource(db_conn, seed=420)
    datasource.load_data(sql_query="SELECT * FROM dummy_data")

    schema = BitfountSchema()
    schema.add_datasource_tables(datasource, table_name="dummy_data")
    schema.add_datasource_tables(datasource, table_name="dummy_data_2")

    datastructure = DataStructure(
        target=target, ignore_cols=["image"], table="dummy_data"
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _BitfountDataset(
        datasource=datasource,
        target=target,
        selected_cols=datastructure.selected_cols,
        data_split=DataSplit.TRAIN,
        schema=schema.get_table_schema("dummy_data"),
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
    )


@fixture
def iter_tabular_dataset(dataframe: pd.DataFrame) -> _IterableBitfountDataset:
    """Basic tabular dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe, ignore_cols=["image"])
    datasource.load_data()

    schema = BitfountSchema()
    schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target, ignore_cols=["image"], table=TABLE_NAME
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _IterableBitfountDataset(
        datasource=datasource,
        target=target,
        selected_cols=datastructure.selected_cols,
        data_split=DataSplit.TRAIN,
        schema=schema.get_table_schema(TABLE_NAME),
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
    )


@fixture
def iter_db_dataset(
    db_session: sqlalchemy.engine.base.Engine,
) -> _IterableBitfountDataset:
    """Basic tabular dataset for tests as fixture."""
    target = "TARGET"
    db_conn = DatabaseConnection(db_session, table_names=["dummy_data", "dummy_data_2"])
    datasource = DatabaseSource(db_conn, seed=420)
    datasource.load_data()

    schema = BitfountSchema()
    schema.add_datasource_tables(datasource, table_name="dummy_data")
    schema.add_datasource_tables(datasource, table_name="dummy_data_2")

    datastructure = DataStructure(
        target=target, ignore_cols=["image"], table="dummy_data"
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _IterableBitfountDataset(
        datasource=datasource,
        target=target,
        selected_cols=datastructure.selected_cols,
        data_split=DataSplit.TRAIN,
        schema=schema.get_table_schema("dummy_data"),
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
    )


@fixture
def image_dataset(dataframe: pd.DataFrame) -> _BitfountDataset:
    """Basic image dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe)
    datasource.load_data()
    schema = BitfountSchema(
        datasource,
        force_stypes={TABLE_NAME: {"image": ["image"]}},
        table_name=TABLE_NAME,
    )

    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target,
        table=TABLE_NAME,
        selected_cols=["image", target],
        image_cols=["image"],
        batch_transforms=[
            {
                "albumentations": {
                    "step": "train",
                    "output": True,
                    "arg": "image",
                    "transformations": [
                        {"Resize": {"height": 224, "width": 224}},
                        "Normalize",
                    ],
                }
            }
        ],
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _BitfountDataset(
        datasource=datasource,
        target=target,
        schema=schema.get_table_schema(TABLE_NAME),
        selected_cols=datastructure.selected_cols,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        batch_transforms=datastructure.get_batch_transformations(),
        data_split=DataSplit.TRAIN,
    )


@fixture
def seg_image_dataset(tmp_path: Path) -> _BitfountDataset:
    """Basic segmentation image dataset for tests as fixture."""
    data = create_segmentation_dataset(seg_dir=tmp_path, count=10)
    dataframe = pd.DataFrame(data)
    target = "masks"
    datasource = DataFrameSource(dataframe)
    datasource.load_data()
    schema = BitfountSchema(
        datasource,
        force_stypes={TABLE_NAME: {"image": ["img", "masks"]}},
        table_name=TABLE_NAME,
    )

    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target,
        table=TABLE_NAME,
        selected_cols=["img", target],
        image_cols=["img", target],
        batch_transforms=[
            {
                "albumentations": {
                    "step": "train",
                    "output": True,
                    "arg": "masks",
                    "transformations": [
                        {"Resize": {"height": 100, "width": 100}},
                        "Normalize",
                    ],
                }
            }
        ],
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _BitfountDataset(
        datasource=datasource,
        target=target,
        schema=schema.get_table_schema(TABLE_NAME),
        selected_cols=datastructure.selected_cols,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        batch_transforms=datastructure.get_batch_transformations(),
        data_split=DataSplit.TRAIN,
    )


@fixture
def image_tab_dataset(dataframe: pd.DataFrame) -> _BitfountDataset:
    """Basic tabular and image dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(dataframe)
    datasource.load_data()
    schema = BitfountSchema()
    schema.add_datasource_tables(
        datasource,
        force_stypes={TABLE_NAME: {"image": ["image"]}},
        table_name=TABLE_NAME,
    )
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(target=target, image_cols=["image"], table=TABLE_NAME)
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])
    return _BitfountDataset(
        datasource=datasource,
        target=target,
        selected_cols=datastructure.selected_cols,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        schema=schema.get_table_schema(TABLE_NAME),
        data_split=DataSplit.TRAIN,
    )


@fixture
def multiimage_dataframe() -> pd.DataFrame:
    """Underlying dataframe for multi-image dataset."""
    return create_dataset(multiimage=True)


@fixture
def multiimage_dataset(multiimage_dataframe: pd.DataFrame) -> _BitfountDataset:
    """Basic multi-image dataset for tests as fixture."""
    target = "TARGET"
    datasource = DataFrameSource(multiimage_dataframe)
    datasource.load_data()
    schema = BitfountSchema()
    schema.add_datasource_tables(
        datasource,
        force_stypes={TABLE_NAME: {"image": ["image1", "image2"]}},
        table_name=TABLE_NAME,
    )
    datasource.data = datasource.data.drop(
        columns=schema.get_feature_names(TABLE_NAME, SemanticType.TEXT)
    )
    datastructure = DataStructure(
        target=target,
        selected_cols=["image1", "image2", target],
        table=TABLE_NAME,
    )
    datastructure.set_training_column_split_by_semantic_type(schema.tables[0])

    return _BitfountDataset(
        datasource=datasource,
        target=target,
        selected_cols=datastructure.selected_cols,
        selected_cols_semantic_types=datastructure.selected_cols_w_types,
        schema=schema.get_table_schema(TABLE_NAME),
        data_split=DataSplit.TRAIN,
    )



"""Custom exceptions for the hub package."""
from bitfount.exceptions import BitfountError


class PodDoesNotExistError(BitfountError):
    """Errors related to references to a non-existent Pod."""

    pass


class SchemaUploadError(BitfountError, ValueError):
    """Could not upload schema to hub."""

    pass


class ModelUploadError(BitfountError):
    """Error occurred whilst uploading model to hub."""

    pass


class ModelValidationError(ModelUploadError):
    """Error occurred in validating model format."""

    pass


class ModelTooLargeError(ModelUploadError, ValueError):
    """The model is too large to upload to the hub."""

    pass



"""Authorisation Checkers."""
from __future__ import annotations

import asyncio
from dataclasses import dataclass
from datetime import datetime, timezone
from enum import Enum
from typing import Callable, Final, List, Mapping, Protocol, Tuple, Type, TypeVar, Union

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
import requests
from requests import HTTPError, RequestException, Response
from requests.exceptions import InvalidJSONError

from bitfount import web_utils
from bitfount.federated.encryption import _RSAEncryption
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.pod_response_message import _PodResponseMessage
from bitfount.federated.task_requests import (
    _EncryptedTaskRequest,
    _SignedEncryptedTaskRequest,
    _TaskRequest,
    _TaskRequestMessage,
)
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.federated.types import (
    SerializedProtocol,
    _PodResponseType,
    _TaskRequestMessageGenerator,
)
from bitfount.hub.api import BitfountAM
from bitfount.hub.authentication_flow import (
    _AUTHORIZATION_PENDING_ERROR,
    _SLOW_DOWN_ERROR,
    _get_auth_environment,
)
from bitfount.hub.types import (
    _DeviceAccessTokenFailResponseJSON,
    _DeviceAccessTokenRequestDict,
    _DeviceAccessTokenResponseJSON,
    _PKCEAccessTokenRequestDict,
    _PKCEAccessTokenResponseJSON,
)

logger = _get_federated_logger(__name__)

# This forces `requests` to make IPv4 connections
# TODO: [BIT-1443] Remove this once Hub/AM support IPv6
requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950


class IdentityVerificationMethod(Enum):
    """Allowed types of identity verification."""

    KEYS = "key-based"
    LOCAL = "local"
    OIDC_ACF_PKCE = "oidc-auth-code"
    OIDC_DEVICE_CODE = "oidc-device-code"
    SAML = "saml"

    DEFAULT = OIDC_DEVICE_CODE


#: Allowed types of identity verification as a tuple of strings
IDENTITY_VERIFICATION_METHODS: Final[Tuple[str, ...]] = tuple(
    i.value for i in IdentityVerificationMethod
)
#: Default identity verification method
DEFAULT_IDENTITY_VERIFICATION_METHOD: Final[
    str
] = IdentityVerificationMethod.DEFAULT.value


def check_identity_verification_method(
    method: Union[str, IdentityVerificationMethod]
) -> IdentityVerificationMethod:
    """Checks that the verification method is supported."""
    try:
        return IdentityVerificationMethod(method)
    except ValueError:
        raise ValueError(f"Unsupported identity verification method: {method}")


_M_co = TypeVar("_M_co", covariant=True)


class _AuthorisationChecker(Protocol[_M_co]):
    """Protocol for checking authorisation.

    This is a simple protocol for checking authorisation. It's used to inform a decision
    on whether a Pod accepts or rejects a job.
    """

    serialized_protocol: SerializedProtocol

    @classmethod
    def create_task_request_message_generator(
        cls,
    ) -> _TaskRequestMessageGenerator:
        """Creates a callable that can be used to generate task request messages."""
        ...

    @classmethod
    def extract_from_task_request_message(
        cls, task_request_message: Union[bytes, _TaskRequestMessage]
    ) -> _M_co:
        """Extract the packed task request details from the task request message.

        Args:
            task_request_message: The task request message, either packed as bytes
                or unpacked.

        Returns:
            The packed task request details, in the dataclass relevant for this
            authorisation checker.
        """
        ...

    @classmethod
    def unpack_task_request(
        cls, msg: Union[bytes, _TaskRequestMessage], pod_private_key: RSAPrivateKey
    ) -> _TaskRequest:
        """Unpacks/decrypts the task request details from the task request message.

        Args:
            msg: The task request message, either packed as bytes or unpacked.
            pod_private_key: The pod's RSA private key, for decryption.
        """
        ...

    async def check_authorisation(self) -> _PodResponseMessage:
        """Check if the task should be executed or not."""
        ...


def _make_encrypted_task_request_message(
    auth_type: IdentityVerificationMethod,
    serialized_protocol: SerializedProtocol,
    pod_identifiers: List[str],
    aes_key: bytes,
    pod_public_key: RSAPublicKey,
) -> bytes:
    """Common function for generating messages with _EncryptedTaskRequest contents."""
    # Create task request
    tr = _TaskRequest(serialized_protocol, pod_identifiers, aes_key)

    # Encrypt and pack in class
    packed_tr = _EncryptedTaskRequest(
        _RSAEncryption.encrypt(tr.serialize(), pod_public_key)
    )

    # Create outer message
    message = _TaskRequestMessage(
        serialized_protocol,
        auth_type.value,
        packed_tr.serialize(),
    )
    return message.serialize()


def _extract_encrypted_task_request(
    task_request_message: Union[bytes, _TaskRequestMessage]
) -> _EncryptedTaskRequest:
    """Extract _EncryptedTaskRequest from _TaskRequestMessage."""
    # Deserialize outer message if provided as bytes
    if isinstance(task_request_message, bytes):
        task_request_message = _TaskRequestMessage.deserialize(task_request_message)

    # Deserialize inner _EncryptedTaskRequest
    packed_task_request = _EncryptedTaskRequest.deserialize(
        task_request_message.request
    )

    return packed_task_request


def _unpack_task_request(
    msg: Union[bytes, _TaskRequestMessage],
    pod_private_key: RSAPrivateKey,
    extract_from_task_request_message: Union[
        Callable[[Union[bytes, _TaskRequestMessage]], _EncryptedTaskRequest],
        Callable[[Union[bytes, _TaskRequestMessage]], _SignedEncryptedTaskRequest],
    ],
) -> _TaskRequest:
    """Function for common behaviour to fully unpack _TaskRequest message.

    Args:
        msg: The _TaskRequestMessage, maybe serialized as bytes.
        pod_private_key: The RSA private key of the pod.
        extract_from_task_request_message:
            Callable that will extract the packed task request from the
            _TaskRequestMessage.

    Returns:
        The unpacked task request.
    """
    if isinstance(msg, bytes):
        msg = _TaskRequestMessage.deserialize(msg)
    packed_task_request = extract_from_task_request_message(msg)
    task_request = _TaskRequest.deserialize(
        _RSAEncryption.decrypt(packed_task_request.encrypted_request, pod_private_key)
    )
    return task_request


@dataclass
class _LocalAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
    """Local Authorisation - Accepts anything.

    When running all pods locally, we can assume you're happy
    and that anything you're trying to run is accepted.

    NB: this is currently not used in production.
    """

    pod_response_message: _PodResponseMessage
    serialized_protocol: SerializedProtocol

    @classmethod
    def _generate_task_request_message(
        cls,
        serialized_protocol: SerializedProtocol,
        pod_identifiers: List[str],
        aes_key: bytes,
        pod_public_key: RSAPublicKey,
    ) -> bytes:
        return _make_encrypted_task_request_message(
            IdentityVerificationMethod.LOCAL,
            serialized_protocol,
            pod_identifiers,
            aes_key,
            pod_public_key,
        )

    @classmethod
    def extract_from_task_request_message(
        cls, task_request_message: Union[bytes, _TaskRequestMessage]
    ) -> _EncryptedTaskRequest:
        return _extract_encrypted_task_request(task_request_message)

    @classmethod
    def create_task_request_message_generator(
        cls,
    ) -> _TaskRequestMessageGenerator:
        return cls._generate_task_request_message

    @classmethod
    def unpack_task_request(
        cls, msg: Union[bytes, _TaskRequestMessage], pod_private_key: RSAPrivateKey
    ) -> _TaskRequest:
        return _unpack_task_request(
            msg, pod_private_key, cls.extract_from_task_request_message
        )

    async def check_authorisation(self) -> _PodResponseMessage:
        """Always returns an accept message."""
        return self.pod_response_message


@dataclass
class _SignatureBasedAuthorisation(_AuthorisationChecker[_SignedEncryptedTaskRequest]):
    """Task authorisation using Signatures.

    Checks whether the Modeller provided signature of the task
    can be verified against the modeller's public key.
    """

    pod_response_message: _PodResponseMessage
    access_manager: BitfountAM
    modeller_name: str
    encrypted_task_request: bytes
    signature: bytes
    serialized_protocol: SerializedProtocol

    @dataclass
    class _SignedMessageMaker:
        private_key: RSAPrivateKey

        def generate_task_request_message(
            self,
            serialized_protocol: SerializedProtocol,
            pod_identifiers: List[str],
            aes_key: bytes,
            pod_public_key: RSAPublicKey,
        ) -> bytes:
            tr = _TaskRequest(serialized_protocol, pod_identifiers, aes_key)

            encrypted_tr = _RSAEncryption.encrypt(tr.serialize(), pod_public_key)
            signature = _RSAEncryption.sign_message(self.private_key, encrypted_tr)
            packed_tr = _SignedEncryptedTaskRequest(encrypted_tr, signature)

            message = _TaskRequestMessage(
                serialized_protocol,
                IdentityVerificationMethod.KEYS.value,
                packed_tr.serialize(),
            )
            return message.serialize()

    # This is an intentional LSP violation as this class is (currently) the only
    # one with a special create_task_request_message_generator method. As the base
    # class is a Protocol, it cannot be instantiated (so an end-user won't
    # accidentally end up with a different create_task_request_message_generator
    # signature) and for the cases where we need to call this version of the method
    # we can assuage static typing with an isinstance or cast against this class.
    #
    # If more classes appear with different create_task_request_message_generator
    # signatures then we should split the signatures out into their own protocols.
    @classmethod
    def create_task_request_message_generator(  # type: ignore[override] # Reason: see comment # noqa: B950
        cls,
        modeller_private_key: RSAPrivateKey,
    ) -> _TaskRequestMessageGenerator:
        return cls._SignedMessageMaker(
            modeller_private_key
        ).generate_task_request_message

    @classmethod
    def extract_from_task_request_message(
        cls, task_request_message: Union[bytes, _TaskRequestMessage]
    ) -> _SignedEncryptedTaskRequest:
        if isinstance(task_request_message, bytes):
            task_request_message = _TaskRequestMessage.deserialize(task_request_message)
        packed_task_request = _SignedEncryptedTaskRequest.deserialize(
            task_request_message.request
        )
        return packed_task_request

    @classmethod
    def unpack_task_request(
        cls, msg: Union[bytes, _TaskRequestMessage], pod_private_key: RSAPrivateKey
    ) -> _TaskRequest:
        return _unpack_task_request(
            msg, pod_private_key, cls.extract_from_task_request_message
        )

    def _check_access(self) -> _PodResponseType:
        """Check that access is allowed for the authorised user and request."""
        return self.access_manager.check_signature_based_access_request(
            unsigned_task=self.encrypted_task_request,
            signed_task=_SignedEncryptedTaskRequest(
                self.encrypted_task_request, self.signature
            ).serialize(),
            pod_identifier=self.pod_response_message.pod_identifier,
            serialized_protocol=self.serialized_protocol,
            modeller_name=self.modeller_name,
        )

    async def check_authorisation(self) -> _PodResponseMessage:
        """Check signature matches modeller public key."""
        self.pod_response_message.add(self._check_access())
        return self.pod_response_message


@dataclass
class _SAMLAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
    """SAML Flow for authorisation.

    This performs a SAML flow,
    sending a SAML challenge and awaiting a response
    then verifying the response with the access manager
    the task is also sent to the access manager to ensure
    that the permissions match up as well.
    """

    pod_response_message: _PodResponseMessage
    access_manager: BitfountAM
    mailbox: _WorkerMailbox
    serialized_protocol: SerializedProtocol

    @classmethod
    def _generate_task_request_message(
        cls,
        serialized_protocol: SerializedProtocol,
        pod_identifiers: List[str],
        aes_key: bytes,
        pod_public_key: RSAPublicKey,
    ) -> bytes:
        return _make_encrypted_task_request_message(
            IdentityVerificationMethod.SAML,
            serialized_protocol,
            pod_identifiers,
            aes_key,
            pod_public_key,
        )

    @classmethod
    def create_task_request_message_generator(
        cls,
    ) -> _TaskRequestMessageGenerator:
        return cls._generate_task_request_message

    @classmethod
    def extract_from_task_request_message(
        cls, task_request_message: Union[bytes, _TaskRequestMessage]
    ) -> _EncryptedTaskRequest:
        return _extract_encrypted_task_request(task_request_message)

    @classmethod
    def unpack_task_request(
        cls, msg: Union[bytes, _TaskRequestMessage], pod_private_key: RSAPrivateKey
    ) -> _TaskRequest:
        return _unpack_task_request(
            msg, pod_private_key, cls.extract_from_task_request_message
        )

    async def check_authorisation(self) -> _PodResponseMessage:
        """Perform SAML dance.

        This obtains a SAML challenge from the Access Manager,
        forwards it on to the modeller,
        awaits their SAML response (which comes from their IDP)
        and then sends the SAML response to the Access Manager
        so that the identity can be verified as well as checking
        that the Task sent by the modeller is authorised.
        """
        # We keep the SAML Request ID on behalf of the access manager
        # This allows the access manager some confidence that the SAML
        # response isn't unprompted, as only this pod, on this specific
        # task mailbox can handle the outstanding SAML challenge
        # The Access Manager will verify that the request ID in the
        # response from the modeller matches this request ID which never
        # leaves the pod.
        saml_request, saml_request_id = self.access_manager.get_saml_challenge()

        await self.mailbox.issue_saml_challenge(saml_request)

        saml_response = await self.mailbox.get_saml_response()

        # Check with Access Manager that user matches expected identity
        # and that the user has the correct permissions for the task
        # that they have requested.
        authentication_result: _PodResponseType = (
            self.access_manager.validate_saml_response(
                saml_response,
                saml_request_id,
                self.pod_response_message.pod_identifier,
                self.mailbox.modeller_name,
                self.serialized_protocol,
            )
        )
        logger.info(
            f"SAML Exchange complete. "
            f"Will inform modeller: {authentication_result.name}"
        )
        self.pod_response_message.add(authentication_result)
        return self.pod_response_message


@dataclass
class _OIDCAuthorisationCode(_AuthorisationChecker[_EncryptedTaskRequest]):
    """OIDC Authorisation Code Flow with PKCE.

    See: https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow-with-proof-key-for-code-exchange-pkce  # noqa: B950
    """

    pod_response_message: _PodResponseMessage
    access_manager: BitfountAM
    mailbox: _WorkerMailbox
    serialized_protocol: SerializedProtocol

    # Auth-related params
    _auth_domain: str = None  # type: ignore[assignment] # Reason: filled in __post_init__ if not provided # noqa: B950
    _client_id: str = None  # type: ignore[assignment] # Reason: filled in __post_init__ if not provided # noqa: B950

    def __post_init__(self) -> None:
        # Fill auth-related attributes if not provided
        auth_env = _get_auth_environment()
        if not self._auth_domain:
            self._auth_domain = auth_env.auth_domain
        if not self._client_id:
            self._client_id = auth_env.client_id

        self._token_endpoint: str = f"https://{self._auth_domain}/oauth/token"

    @classmethod
    def _generate_task_request_message(
        cls,
        serialized_protocol: SerializedProtocol,
        pod_identifiers: List[str],
        aes_key: bytes,
        pod_public_key: RSAPublicKey,
    ) -> bytes:
        return _make_encrypted_task_request_message(
            IdentityVerificationMethod.OIDC_ACF_PKCE,
            serialized_protocol,
            pod_identifiers,
            aes_key,
            pod_public_key,
        )

    @classmethod
    def create_task_request_message_generator(
        cls,
    ) -> _TaskRequestMessageGenerator:
        return cls._generate_task_request_message

    @classmethod
    def extract_from_task_request_message(
        cls, task_request_message: Union[bytes, _TaskRequestMessage]
    ) -> _EncryptedTaskRequest:
        return _extract_encrypted_task_request(task_request_message)

    @classmethod
    def unpack_task_request(
        cls, msg: Union[bytes, _TaskRequestMessage], pod_private_key: RSAPrivateKey
    ) -> _TaskRequest:
        return _unpack_task_request(
            msg, pod_private_key, cls.extract_from_task_request_message
        )

    async def check_authorisation(self) -> _PodResponseMessage:
        """Perform OIDC flow from the pod-side.

        This flow is based on:
        https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow-with-proof-key-for-code-exchange-pkce  # noqa: B950
        """
        # Retrieve client ID for pod and send to modeller
        client_id = self._get_client_id()
        await self.mailbox.send_oidc_client_id(client_id)

        # Wait for authorization code and code verifier
        response = await self.mailbox.get_oidc_auth_flow_response()

        # Get access token
        modeller_access_token = self._get_access_token(
            response.auth_code, response.code_verifier, response.redirect_uri
        )

        # Check with Access Manager that user matches expected identity
        # and that the user has the correct permissions for the task
        # that they have requested.
        authentication_result: _PodResponseType = self._check_access(
            modeller_access_token
        )

        logger.info(
            f"OIDC authorisation check complete. "
            f"Will inform modeller: {authentication_result.name}"
        )
        self.pod_response_message.add(authentication_result)
        return self.pod_response_message

    def _get_client_id(self) -> str:
        """Retrieve client ID for pod."""
        # Note: Currently this just returns the static client ID used throughout.
        #       If we change to using Dynamic Client Registration (https://auth0.com/docs/api/authentication#dynamic-application-client-registration)  # noqa: B950
        #       we should change what is returned by this method
        return self._client_id

    def _get_access_token(
        self,
        auth_code: str,
        code_verifier: str,
        redirect_uri: str,
    ) -> str:
        """Request access token from oauth endpoint."""
        # POST to endpoint to exchange for access token.
        # See: https://auth0.com/docs/api/authentication?http#authorization-code-flow-with-pkce45  # noqa: B950
        request_data: _PKCEAccessTokenRequestDict = {
            "grant_type": "authorization_code",
            "client_id": self._client_id,
            "code": auth_code,
            "code_verifier": code_verifier,
            "redirect_uri": redirect_uri,
        }
        response: Response = web_utils.post(
            self._token_endpoint,
            data=request_data,
            timeout=20,
        )
        response.raise_for_status()

        # Handle good response
        response_json: _PKCEAccessTokenResponseJSON = response.json()
        return response_json["access_token"]

    def _check_access(self, modeller_access_token: str) -> _PodResponseType:
        """Check that access is allowed for the authorised user and request."""
        return self.access_manager.check_oidc_access_request(
            self.pod_response_message.pod_identifier,
            self.serialized_protocol,
            self.mailbox.modeller_name,
            modeller_access_token,
        )


@dataclass
class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
    """OIDC Device Authorisation Flow from pod-side.

    See: https://auth0.com/docs/get-started/authentication-and-authorization-flow/device-authorization-flow  # noqa: B950
    """

    pod_response_message: _PodResponseMessage
    access_manager: BitfountAM
    mailbox: _WorkerMailbox
    serialized_protocol: SerializedProtocol

    # Auth-related params
    _auth_domain: str = None  # type: ignore[assignment] # Reason: filled in __post_init__ if not provided # noqa: B950
    _client_id: str = None  # type: ignore[assignment] # Reason: filled in __post_init__ if not provided # noqa: B950

    def __post_init__(self) -> None:
        # Fill auth-related attributes if not provided
        auth_env = _get_auth_environment()
        if not self._auth_domain:
            self._auth_domain = auth_env.auth_domain
        if not self._client_id:
            self._client_id = auth_env.client_id

        self._token_endpoint: str = f"https://{self._auth_domain}/oauth/token"

    @classmethod
    def _generate_task_request_message(
        cls,
        serialized_protocol: SerializedProtocol,
        pod_identifiers: List[str],
        aes_key: bytes,
        pod_public_key: RSAPublicKey,
    ) -> bytes:
        return _make_encrypted_task_request_message(
            IdentityVerificationMethod.OIDC_DEVICE_CODE,
            serialized_protocol,
            pod_identifiers,
            aes_key,
            pod_public_key,
        )

    @classmethod
    def create_task_request_message_generator(
        cls,
    ) -> _TaskRequestMessageGenerator:
        """See parent class for more information."""
        return cls._generate_task_request_message

    @classmethod
    def extract_from_task_request_message(
        cls, task_request_message: Union[bytes, _TaskRequestMessage]
    ) -> _EncryptedTaskRequest:
        """See parent class for more information."""
        return _extract_encrypted_task_request(task_request_message)

    @classmethod
    def unpack_task_request(
        cls, msg: Union[bytes, _TaskRequestMessage], pod_private_key: RSAPrivateKey
    ) -> _TaskRequest:
        """See parent class for more information."""
        return _unpack_task_request(
            msg, pod_private_key, cls.extract_from_task_request_message
        )

    async def check_authorisation(self) -> _PodResponseMessage:
        """Perform OIDC flow from the pod-side.

        This flow is based on:
        https://auth0.com/docs/get-started/authentication-and-authorization-flow/device-authorization-flow  # noqa: B950
        """
        # Send client ID for pod to modeller
        logger.info("Sending client ID for pod to modeller")
        await self.mailbox.send_oidc_client_id(self._client_id)

        # Wait for device code from modeller
        logger.info("Awaiting identity verification details from modeller")
        device_code_details = await self.mailbox.get_oidc_device_code_response()

        # Poll for user approval/authentication
        logger.info("Awaiting modeller approval for identity verification access")
        modeller_access_token = await self._poll_for_access_token(
            device_code_details.device_code,
            device_code_details.expires_at,
            device_code_details.interval,
        )

        # Check with Access Manager that user matches expected identity
        # and that the user has the correct permissions for the task
        # that they have requested.
        logger.info("Verifying access permissions for modeller task")
        authentication_result: _PodResponseType = self._check_access(
            modeller_access_token
        )

        logger.info(
            f"OIDC authorisation check complete. "
            f"Will inform modeller: {authentication_result.name}"
        )
        self.pod_response_message.add(authentication_result)
        return self.pod_response_message

    async def _poll_for_access_token(
        self, device_code: str, expires_at: datetime, interval: int
    ) -> str:
        """Polls the token endpoint waiting for the authenticated response.

        Polls at the interval specified (which comes from the modeller's
        """
        # Poll until the token has expired
        while self._has_not_expired(expires_at):
            # Make request
            request_data: _DeviceAccessTokenRequestDict = {
                "client_id": self._client_id,
                "device_code": device_code,
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            }
            response = web_utils.post(
                self._token_endpoint,
                data=request_data,
                timeout=10,
            )

            # If we get our access token, exit early
            if (status_code := response.status_code) == 200:
                try:
                    success_json: _DeviceAccessTokenResponseJSON = response.json()
                    return success_json["access_token"]
                except (InvalidJSONError, KeyError):
                    raise RequestException(
                        f'Success response, but JSON was invalid: "{response.text}"'
                    )

            # Otherwise, treat as a retry-able error until we know otherwise;
            # status code could be any 4XX value, so we instead just check for the
            # right format and error values.
            try:
                error_json: _DeviceAccessTokenFailResponseJSON = response.json()
            except InvalidJSONError:
                raise RequestException(
                    f'{status_code} response, but JSON was invalid: "{response.text}"'
                )

            # See if just pending authorisation
            if (error := error_json.get("error")) == _AUTHORIZATION_PENDING_ERROR:
                logger.info(
                    "Waiting on modeller to approve identity verification request..."
                )
            # We're requesting too quickly, increase interval
            elif error == _SLOW_DOWN_ERROR:
                logger.warning(
                    f"Polling token endpoint too frequently. "
                    f"Increasing interval from {interval} to {interval+1}"
                )
                interval += 1

            # Otherwise, it's an actual error, raise it
            else:
                # If we have an actual error message, raise that
                if error:
                    error_msg = f"Error in OAuth response ({status_code}) ({error})"
                    # Try to add error_description if present
                    try:
                        error_msg += f": {error_json['error_description']}"
                    except KeyError:
                        pass
                    raise HTTPError(error_msg, response=response)

                # Otherwise, raise an exception with as much detail as possible
                else:
                    raise HTTPError(
                        f"An unexpected error occurred: "
                        f'status code: {status_code}; "{response.text}"'
                    )

            # Sleep for the interval
            await asyncio.sleep(interval)

        # If we're past the expiration, raise it
        raise TimeoutError("Device code has expired, unable to retrieve access token.")

    @staticmethod
    def _has_not_expired(expires_at: datetime) -> bool:
        return datetime.now(timezone.utc) < expires_at

    def _check_access(self, modeller_access_token: str) -> _PodResponseType:
        """Check that access is allowed for the authorised user and request."""
        return self.access_manager.check_oidc_access_request(
            self.pod_response_message.pod_identifier,
            self.serialized_protocol,
            self.mailbox.modeller_name,
            modeller_access_token,
        )


_IDENTITY_VERIFICATION_METHODS_MAP: Final[
    Mapping[IdentityVerificationMethod, Type[_AuthorisationChecker]]
] = {
    IdentityVerificationMethod.KEYS: _SignatureBasedAuthorisation,
    IdentityVerificationMethod.LOCAL: _LocalAuthorisation,
    IdentityVerificationMethod.OIDC_ACF_PKCE: _OIDCAuthorisationCode,
    IdentityVerificationMethod.OIDC_DEVICE_CODE: _OIDCDeviceCode,
    IdentityVerificationMethod.SAML: _SAMLAuthorisation,
}



"""Operations done at batch time defined here."""
from __future__ import annotations

import inspect
from typing import Any, Dict, List, Protocol, Type, Union, cast

import albumentations as A
import albumentations.augmentations as albumentations_augmentations  # type: ignore[import] # Reason: only used for vars extraction, don't need type hints # noqa: B950
import albumentations.pytorch as albumentations_pytorch  # type: ignore[import] # Reason: only used for vars extraction, don't need type hints # noqa: B950
import attr
from marshmallow import fields, post_load
from marshmallow_union import Union as M_Union
import numpy as np

from bitfount.config import _PYTORCH_ENGINE, BITFOUNT_ENGINE
from bitfount.data.types import DataSplit
from bitfount.transformations.base_transformation import _TransformationSchema
from bitfount.transformations.exceptions import TransformationParsingError
from bitfount.transformations.unary_operations import UnaryOperation
from bitfount.types import _JSONDict

#: Dictionary of available image transformations and their corresponding classes.
from bitfount.utils import delegates

IMAGE_TRANSFORMATIONS: Dict[str, Type[A.BasicTransform]] = {
    name: class_
    for name, class_ in vars(albumentations_augmentations).items()
    if inspect.isclass(class_)
    and issubclass(class_, A.BasicTransform)
    and not inspect.isabstract(class_)
}
if BITFOUNT_ENGINE == _PYTORCH_ENGINE:
    IMAGE_TRANSFORMATIONS.update(
        {
            name: class_
            for name, class_ in vars(albumentations_pytorch).items()
            if inspect.isclass(class_)
            and issubclass(class_, A.BasicTransform)
            and not inspect.isabstract(class_)
        }
    )


@delegates()
@attr.dataclass(kw_only=True)
class BatchTimeOperation(UnaryOperation):
    """Class just to denote that transformation will happen at batch time.

    All batch time operations must be unary operations.

    Args:
        step: Denotes whether transformations should be performed at training,
            validation or test time.

    """

    step: DataSplit


@delegates()
@attr.dataclass(kw_only=True)
class AlbumentationsImageTransformation(BatchTimeOperation):
    """Represents image transformations done on a single column at batch time.

    Args:
        transformations: List of transformations to be performed in order as one
            transformation.

    Raises:
        ValueError: If the `output` is set to False.
    """

    _registry_name = "albumentations"

    transformations: List[Union[str, Dict[str, _JSONDict]]]

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        if not self.output:
            raise ValueError("`output` cannot be False for a BatchTimeOperation")

    def _load_transformation(
        self, tfm: Union[str, Dict[str, _JSONDict]]
    ) -> A.BasicTransform:
        """Loads and returns transformation in albumentations.

        Args:
            tfm: Albumentations transformation.

        Raises:
            TransformationParsingError: If the transformation name cannot be parsed
                properly.
            ValueError: If the transformation cannot be found.

        Returns:
            The transform loaded in albumentations.
        """
        if isinstance(tfm, dict):
            tfm_keys = list(tfm.keys())
            if len(tfm_keys) > 1:
                raise TransformationParsingError(
                    f"Transformation has supplied multiple names: {tfm_keys}"
                )
            tfm_name = tfm_keys[0]
            tfm_args = tfm[tfm_name]
        else:
            tfm_name = tfm
            tfm_args = {}
        transformation = IMAGE_TRANSFORMATIONS.get(tfm_name)
        if transformation is None:
            raise ValueError(f"Transformation {tfm_name} could not be found.")

        return transformation(**tfm_args)

    def get_callable(self) -> _AlbumentationsAugmentation:
        """Returns callable which performs the transformations.

        Returns:
            The callable to perform transformations.
        """
        list_of_transformations: List[A.BasicTransform] = []
        for tfm in self.transformations:
            a_tfm = self._load_transformation(tfm)
            list_of_transformations.append(a_tfm)

        tfm_callable = A.Compose(list_of_transformations)
        return cast(_AlbumentationsAugmentation, tfm_callable)

    class _Schema(_TransformationSchema):
        """Marshmallow schema for ImageTransformation."""

        # From Transformation
        name = fields.String(default=None)
        output = fields.Boolean(default=True)
        # From UnaryOperation
        arg = fields.String(required=True)
        # From BatchTimeOperation
        step = fields.Method(
            serialize="_serialize_step", deserialize="_deserialize_step"
        )
        # From ImageTransformation
        transformations = fields.List(
            M_Union(
                [
                    fields.String(),
                    fields.Dict(
                        keys=fields.String(),
                        values=fields.Dict(keys=fields.String(), values=fields.Raw()),
                    ),
                ],
            ),
            required=True,
        )

        @staticmethod
        def _serialize_step(transformation: AlbumentationsImageTransformation) -> str:
            """Serializes the step of the transformation from Enum to string."""
            return transformation.step.value

        @staticmethod
        def _deserialize_step(value: str) -> DataSplit:
            """Deserializes the step of the transformation from string to Enum."""
            return DataSplit(value)

        @post_load
        def make_transformation(
            self, data: _JSONDict, **_kwargs: Any
        ) -> AlbumentationsImageTransformation:
            """Creates the ImageTransformation object from the marshmallow data."""
            return AlbumentationsImageTransformation(**data)


class _AlbumentationsAugmentation(Protocol):
    """Protocol for the signature of an albumentations augmentation function."""

    def __call__(self, *, image: np.ndarray) -> Dict[str, np.ndarray]:
        """Calls the function.

        `image` must be passed as a kwarg.
        """
        ...



"""Private SQL query algorithm."""
from __future__ import annotations

import math
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Final,
    List,
    Mapping,
    Optional,
    Tuple,
    cast,
)

from marshmallow import fields
import pandas as pd
import snsql
from snsql import Privacy
from sqlalchemy import inspect
import sqlparse

from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.exceptions import DatabaseSchemaNotFoundError
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import SemanticType, _SemanticTypeValue
from bitfount.federated.algorithms.base import (
    _BaseAlgorithmFactory,
    _BaseModellerAlgorithm,
    _BaseWorkerAlgorithm,
)
from bitfount.federated.exceptions import PrivateSqlError
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.mixins import _ModellessAlgorithmMixIn
from bitfount.federated.privacy.differential import (
    DPModellerConfig,
    DPPodConfig,
    _DifferentiallyPrivate,
)
from bitfount.types import T_FIELDS_DICT
from bitfount.utils import delegates

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub

logger = _get_federated_logger(__name__)


class _ModellerSide(_BaseModellerAlgorithm):
    """Modeller side of the PrivateSqlQuery algorithm."""

    def initialise(
        self,
        **kwargs: Any,
    ) -> None:
        """Nothing to initialise here."""
        pass

    def run(self, results: Mapping[str, pd.DataFrame]) -> Dict[str, pd.DataFrame]:
        """Simply returns results."""
        return dict(results)


class _WorkerSide(_BaseWorkerAlgorithm):
    """Worker side of the PrivateSqlQuery algorithm."""

    _SMART_NOISE_TYPE: Final[Mapping[str, Tuple[str, Optional[int]]]] = {
        "int": ("int", 1_000),
        "int32": ("int", 1_000),
        "int64": ("int", 1_000),
        "float": ("float", 1),
        "float32": ("float", 1),
        "float64": ("float", 1),
        "string": ("string", None),
    }

    def __init__(
        self,
        *,
        query: str,
        epsilon: float,
        delta: float,
        column_ranges: dict,
        hub: BitfountHub,
        db_schema: Optional[str] = None,
        table: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.datasource: BaseSource
        self.pod_identifier: Optional[str] = None
        self.pod_dp: Optional[DPPodConfig] = None
        self.query = query
        self.epsilon = epsilon
        self.delta = delta
        self.column_ranges = column_ranges
        self.hub = hub
        self.table = table
        self.db_schema = db_schema
        super().__init__(**kwargs)

    def initialise(
        self,
        datasource: BaseSource,
        pod_dp: Optional[DPPodConfig] = None,
        pod_identifier: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Sets Datasource."""
        datasource.load_data()
        self.datasource = datasource
        if pod_identifier:
            self.pod_identifier = pod_identifier
        if pod_dp:
            self.pod_dp = pod_dp

        if (
            self.table is None
            and self.pod_identifier is not None
            and not self.datasource.multi_table
        ):
            self.table = self.pod_identifier.split("/")[1]

    def map_types(self, schema: BitfountSchema) -> None:
        """Map from Pandas DataTypes to SmartNoise types.

        For the features that were user-specified in the single-table dataset,
        map from the Pandas DataTypes to the equivalent SmartNoise types.
        """
        for column in self.column_ranges:
            mapped_type = None
            default_max = None
            for table_name in schema.table_names:
                # Determine what type we have in the schema for the column
                for stype in SemanticType:
                    for feature in schema.get_table_schema(
                        table_name
                    ).get_feature_names(stype):
                        if feature == column:
                            mapped_type = (
                                schema.get_table_schema(table_name)
                                .features[cast(_SemanticTypeValue, stype.value)][
                                    feature
                                ]
                                .dtype.name
                            )
                            break

            if mapped_type is None:
                logger.warning(
                    f"No field named '{column}' present in the schema"
                    "will proceed assuming it is a string."
                )
                mapped_type = "str"

            # Map the type we have to an equivalent for SmartNoise SQL
            sn_mapped_type, default_max = self._SMART_NOISE_TYPE.get(
                mapped_type.lower(), (None, None)
            )
            if not sn_mapped_type:
                logger.warning(
                    f"Type {mapped_type} for column '{column}' is not supported"
                    "defaulting to string type."
                )
                sn_mapped_type = "string"

            self.column_ranges[column]["type"] = sn_mapped_type
            # if there is no specified lower/upper for int/float set defaults
            if "lower" not in self.column_ranges[column] and default_max is not None:
                logger.warning(
                    "Using default lower and upper bound of "
                    f"(0, {default_max}) for field [{column}]"
                )
                self.column_ranges[column]["lower"] = 0
                self.column_ranges[column]["upper"] = default_max

    def map_types_multi(self, schema: BitfountSchema) -> List[str]:
        """Map from Pandas DataTypes to SmartNoise types.

        For the features that were user-specified in the multi-table dataset,
        map from the Pandas DataTypes to the equivalent SmartNoise types.
        """
        features_added = []
        for table_name, columns in self.column_ranges.items():
            for column in columns:
                mapped_type = None
                default_max = None
                if table_name in schema.table_names:
                    # Determine what type we have in the schema for the column
                    for stype in SemanticType:
                        for feature in schema.get_table_schema(
                            table_name
                        ).get_feature_names(stype):
                            if feature == column:
                                mapped_type = (
                                    schema.get_table_schema(table_name)
                                    .features[cast(_SemanticTypeValue, stype.value)][
                                        feature
                                    ]
                                    .dtype.name
                                )
                                break
                else:
                    logger.warning(
                        f"No table named '{table_name}' present in the schema. "
                        "SQL query will probably fail if a query attempts "
                        "to use this table."
                    )

                if mapped_type is None:
                    logger.warning(
                        f"No field named '{column}' present in the schema"
                        "will proceed assuming it is a string."
                    )
                    mapped_type = "str"

                # Map the type we have to an equivalent for SmartNoise SQL
                sn_mapped_type, default_max = self._SMART_NOISE_TYPE.get(
                    mapped_type.lower(), (None, None)
                )
                if not sn_mapped_type:
                    logger.warning(
                        f"Type {mapped_type} for column '{column}' is not supported"
                        "defaulting to string type."
                    )
                    sn_mapped_type = "string"

                self.column_ranges[table_name][column]["type"] = sn_mapped_type
                features_added.append(column)
                # if there is no specified lower/upper for int/float set defaults
                if (
                    "lower" not in self.column_ranges[table_name][column]
                    and default_max is not None
                ):
                    logger.warning(
                        "Using default lower and upper bound of "
                        f"(0, {default_max}) for field [{column}]"
                    )
                    self.column_ranges[table_name][column]["lower"] = 0
                    self.column_ranges[table_name][column]["upper"] = default_max

        return features_added

    def extend_types(self, schema: BitfountSchema, already_added: List[str]) -> None:
        """Add SmartNoise types for remaining features from Pandas DataTypes.

        In this case, for the features are not all specified in the modeller-supplied
        metadata, map to the appropriate type.
        """
        for table_name in schema.table_names:
            for stype in SemanticType:
                for feature in schema.get_table_schema(table_name).get_feature_names(
                    stype
                ):
                    f_cat = str(stype.value)
                    if feature not in already_added:
                        default_max = None
                        mapped_type = (
                            schema.get_table_schema(table_name)
                            .features[
                                f_cat  # type: ignore[literal-required] # Reason: below
                            ][feature.replace('"', "")]
                            .dtype.name
                        )
                        # Reason: enum value is valid
                        # Only need to deal with numeric types, others already string
                        mapped_type, default_max = self._SMART_NOISE_TYPE.get(
                            mapped_type.lower(), (None, None)
                        )

                        if table_name not in self.column_ranges:
                            self.column_ranges[table_name] = {}
                        self.column_ranges[table_name][feature] = {"type": mapped_type}
                        # If there is no specified lower/upper for int/float, default
                        if default_max is not None:
                            logger.warning(
                                "Using default lower and upper bound of "
                                f"(0,{default_max}) for field [{feature}]"
                            )
                            self.column_ranges[table_name][feature].update({"lower": 0})
                            self.column_ranges[table_name][feature].update(
                                {"upper": default_max}
                            )

    def run(self) -> pd.DataFrame:
        """Executes the SQL query on the `BaseSource`."""
        logger.info("Executing query...")
        query = self.query
        if self.pod_identifier is None:
            raise ValueError("No pod identifier - cannot get schema to infer types.")
        # Get the schema from the hub, needed for getting the column data types.
        schema = self.hub.get_pod_schema(self.pod_identifier)
        # Check that modeller-side dp parameters are within the pod config range.
        dp = _DifferentiallyPrivate({"epsilon": self.epsilon, "delta": self.delta})
        # We initialise dp explicitly above, so dp._dp_config will be of
        # type DPModellerConfig, making it safe to cast.
        dp._dp_config = cast(DPModellerConfig, dp._dp_config)
        if self.pod_dp:
            dp.apply_pod_dp(self.pod_dp)
        else:
            # If there is no pod_pd, we assign to it the epsilon and delta values
            # from the Modeller, as they have been previously checked by the AM.
            self.pod_dp = DPPodConfig(epsilon=self.epsilon, delta=self.delta)

        # Extract the metadata from the column ranges.
        if self.datasource.multi_table and isinstance(self.datasource, DatabaseSource):
            # Check if there is a join in the query and error out appropriately
            parsed = sqlparse.parse(self.query)
            # Get default db schema name to pass to the metadata for snsql
            if self.db_schema is None:
                self.db_schema = inspect(
                    self.datasource.db_conn.con
                ).default_schema_name
            else:
                schemas = inspect(self.datasource.db_conn.con).get_schema_names()
                if self.db_schema not in schemas:
                    raise DatabaseSchemaNotFoundError(
                        f"Schema '{self.db_schema}' not found in database."
                    )
            # If token in query [0] is a keyword, and a join, raise an error
            for tok in parsed[0].tokens:
                if tok.is_keyword:
                    if "JOIN" in tok.value.upper():
                        raise ValueError("JOIN clauses are not supported.")

            # Map the dtypes to types understood by SmartNoise SQL
            already_added = self.map_types_multi(schema)
            # Add the mappings for remaining columns
            self.extend_types(schema, already_added)
            # Set up the metadata dictionary required for SmartNoise SQL
            meta: Dict[str, Dict[str, Dict[str, Dict[str, int]]]] = {
                # "Collection" name
                "Database": {
                    # Pod name
                    self.db_schema: {
                        # "Table" name - added later
                    }
                }
            }
            # We update the metadata with a copy so we don't modify
            # the original column ranges.
            meta["Database"][self.db_schema].update(self.column_ranges.copy())
            for table in meta["Database"][self.db_schema]:
                meta["Database"][self.db_schema][table]["row_privacy"] = False
        else:
            # Table name is set to the pod id for single table datasources
            self.table = cast(str, self.table)
            # Single-table/in-memory path - load the data into memory
            self.datasource.load_data(table_name=self.table)
            df = self.datasource.data
            if (f"from `{self.table}`" not in self.query) and (
                f"FROM `{self.table}`" not in self.query
            ):
                err_msg = """The default table for single table datasource is the pod
                    identifier without the username, in between backticks(``),
                    repeated twice, separated by a dot.
                    (e.g. `census-income-demo`.`census-income-demo`)
                    Please ensure your SQL query operates on that table. The
                    table name should be put inside backticks(``) in the
                    query statement, to make sure it is correctly parsed."""
                logger.error(err_msg)
                raise ValueError(err_msg)

            # We need to remove hyphens as pandasql errors out if
            # they are included in the table name and query.
            aux_query = self.query.split("`")
            table_name = self.table.replace("-", "")
            aux_query[:] = [
                item if item != self.table else table_name for item in aux_query
            ]

            # Update the query with the table name with hyphens removed
            query = "".join(aux_query)

            # Change the table name in the schema for the purpose of this query
            for schema_table in schema.tables:
                if schema_table.name == self.table:
                    schema_table.name = table_name

            # Map the dtypes to types understood by SmartNoise SQL
            self.map_types(schema)
            # Set up the metadata dictionary required for SmartNoise SQL
            meta = {
                "Database": {
                    table_name: {
                        table_name: {"row_privacy": True, "rows": int(len(df.index))}
                    }
                }
            }
            # We update the metadata with a copy so we don't modify
            # the original column ranges.
            meta["Database"][table_name][table_name].update(self.column_ranges.copy())

        # Set default values for the used_epsilon and used_delta
        used_epsilon = float("inf")
        used_delta = float("inf")

        # Initialise epsilon and delta divisors as 1
        delta_div = 1.0
        epsilon_div = 1.0
        # Parameters used for rounding up the divisors
        delta_n = 10
        epsilon_n = 10

        try:
            while used_epsilon > self.pod_dp.epsilon or used_delta > self.pod_dp.delta:
                # Configure privacy
                # Note that the parameters are per-column maximums
                privacy = Privacy(
                    epsilon=dp._dp_config.epsilon / epsilon_div,
                    delta=dp._dp_config.delta / delta_div,
                )
                # Pre-flight the sql query, and compute the privacy cost
                if self.datasource.multi_table and isinstance(
                    self.datasource, DatabaseSource
                ):

                    # Now connect to the database and execute the query.
                    conn = self.datasource.db_conn.con.raw_connection()
                    reader = snsql.from_connection(
                        conn, privacy=privacy, metadata=meta, engine="postgres"
                    )
                    used_epsilon, used_delta = reader.get_privacy_cost(query)

                else:

                    reader = snsql.from_df(df, privacy=privacy, metadata=meta)

                    used_epsilon, used_delta = reader.get_privacy_cost(query)

                if used_epsilon > self.pod_dp.epsilon:
                    # Update epsilon_div to be the result of dividing the used_epsilon
                    # by the maximum allowed value by the pod, rounded up to eps_n
                    # decimals
                    epsilon_div = (
                        math.ceil(used_epsilon / self.pod_dp.epsilon * 10**epsilon_n)
                        / 10**epsilon_n
                    )
                    epsilon_n -= 1

                if used_delta > self.pod_dp.delta and delta_n > 1:
                    # Update delta_div to be the result of dividing the used_delta
                    # by the maximum allowed value by the pod, rounded up to delta_n
                    # decimals
                    delta_div = (
                        math.ceil(used_delta / self.pod_dp.delta * 10**delta_n)
                        / 10**delta_n
                    )
                    delta_n -= 1
                elif used_delta > self.pod_dp.delta and delta_n <= 1:
                    # On the last loop, do a bigger division for the delta.
                    delta_div = (
                        math.ceil(used_delta / self.pod_dp.delta * 10**delta_n)
                        / 10**delta_n
                    ) * 10
                    delta_n -= 1
                # If any of delta_n or epsilon_n are 0, then we have reached
                # the end of our loop, and we raise an error to the user.
                if delta_n < 0:
                    raise PrivateSqlError(
                        "Cannot execute current query given the provided privacy "
                        f"parameter delta={self.delta}). Try again using a "
                        f"smaller value of delta."
                    )
                if epsilon_n < 0:
                    raise PrivateSqlError(
                        "Cannot execute current query given the provided privacy "
                        f"parameter epsilon={self.epsilon}. Try again using a "
                        f"smaller epsilon value."
                    )
            # Log the total values of the privacy budget used.
            logger.federated_info(
                "Total privacy cost for query will be "
                f"(epsilon={used_epsilon}, delta={used_delta})"
            )
            # Log the values of the privacy budget used per column.
            logger.federated_info(
                f"Executing SQL query with epsilon {privacy.epsilon} "
                f"and delta {privacy.delta} applied on each queried column."
            )
            # Execute the Private SQL query
            output = reader.execute(query)

        except Exception as ex:
            raise PrivateSqlError(
                "Error executing PrivateSQL query: " f"[{self.query}], got error [{ex}]"
            )
        return cast(pd.DataFrame, output)


@delegates()
class PrivateSqlQuery(_BaseAlgorithmFactory, _ModellessAlgorithmMixIn):
    """Simple algorithm for running a SQL query on a table, with privacy.

    :::note

    The values provided for the privacy budget (i.e. epsilon and delta)
    will be applied individually to all columns included in the SQL query provided.
    If the total values of the epsilon and delta exceed the maximum allowed by the pod,
    the provided values will be reduced to the maximum values required to remain
    within the allowed privacy budget.

    :::

    Args:
        query: The SQL query to execute.
        epsilon: The maximum epsilon to use for the privacy budget.
        delta: The target delta to use for the privacy budget.
        column_ranges: A dictionary of column names and their ranges.
        table: The target table name. For single table pod datasources,
            this will default to the pod name.
        db_schema: The name of the schema for a database connection. If
            not provided, it will be set to the default schema name
            for the database.

    Attributes:
        query: The SQL query to execute.
        epsilon: The maximum epsilon to use for the privacy budget.
        delta: The target delta to use for the privacy budget.
        column_ranges: A dictionary of column names and their ranges.
        table: The target table name. For single table pod datasources,
            this will default to the pod name.
        db_schema: The name of the schema for a database connection. If
            not provided, it will be set to the default schema name
            for the database.

    Raises:
        PrivateSqlError: If there is an error executing the private SQL query (e.g. DP
            misconfiguration or bad query specified).
        ValueError: If a pod identifier is not supplied, or if a join is attempted.
        DatabaseSchemaNotFoundError: If a non-existent db_schema name is provided.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "query": fields.Str(),
        "epsilon": fields.Float(allow_nan=True),
        "delta": fields.Float(allow_nan=True),
        "column_ranges": fields.Dict(),
        "table": fields.Str(allow_none=True),
        "db_schema": fields.Str(allow_none=True),
    }

    def __init__(
        self,
        *,
        query: str,
        epsilon: float,
        delta: float,
        column_ranges: dict,
        table: Optional[str] = None,
        db_schema: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.query = query
        self.epsilon = epsilon
        self.delta = delta
        self.column_ranges = column_ranges
        self.table = table
        self.db_schema = db_schema

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the PrivateSqlQuery algorithm.

        Args:
            **kwargs: Additional keyword arguments to pass to the modeller side.
        """
        return _ModellerSide(**kwargs)

    def worker(self, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the PrivateSqlQuery algorithm.

        Args:
            **kwargs: Additional keyword arguments to pass to the worker side. `hub`
                must be one of these keyword arguments which provides a `BitfountHub`
                instance.
        """
        return _WorkerSide(
            query=self.query,
            epsilon=self.epsilon,
            delta=self.delta,
            column_ranges=self.column_ranges,
            table=self.table,
            db_schema=self.db_schema,
            **kwargs,
        )



"""Base classes for all algorithms.

Each module in this package defines a single algorithm.

Attributes:
    registry: A read-only dictionary of algorithm factory names to their
              implementation classes.
"""
from __future__ import annotations

from abc import ABC
import inspect
from types import MappingProxyType
from typing import Any, Callable, ClassVar, Dict, Mapping, Optional, Type, TypeVar

from decorator import decorate
from typing_extensions import ParamSpec

from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.privacy.differential import DPPodConfig
from bitfount.federated.roles import _RolesMixIn
from bitfount.federated.types import AlgorithmType
from bitfount.types import T_FIELDS_DICT, T_NESTED_FIELDS, _BaseSerializableObjectMixIn

logger = _get_federated_logger(__name__)

_P = ParamSpec("_P")
_R = TypeVar("_R")


class _BaseAlgorithm(ABC):
    """Blueprint for either the modeller side or the worker side of BaseAlgorithm."""

    def __init__(self, **kwargs: Any):
        super().__init__()

    @classmethod
    def exception_handler(cls, func: Callable[_P, _R]) -> Callable[_P, Optional[_R]]:
        """Log exceptions raised to federated logger."""

        def federated_exception(
            cls: _BaseAlgorithm, *args: _P.args, **kwargs: _P.kwargs
        ) -> Optional[_R]:
            try:
                r = func(*args, **kwargs)
                return r
            except Exception as e:
                # TODO: [BIT-1619] change to federated_exception
                logger.federated_error(str(e))
                return None

        return decorate(func, federated_exception)


class _BaseModellerAlgorithm(_BaseAlgorithm, ABC):
    """Modeller side of the algorithm."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)


class _BaseWorkerAlgorithm(_BaseAlgorithm, ABC):
    """Worker side of the algorithm."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def __init_subclass__(cls, **kwargs: Any):

        if not (inspect.isabstract(cls) or ABC in cls.__bases__):
            if hasattr(cls, "run"):
                # Decorate the `run` method to log exceptions and fail gracefully.
                cls.run = cls.exception_handler(cls.run)  # type: ignore[attr-defined] # reason: decorate all child run methods # noqa: B950
            else:
                raise AttributeError(
                    f"Worker algorithm {cls.__name__} does not have a `run` method;"
                    f" all subclasses should implement a `run()`."
                )
            if hasattr(cls, "initialise"):
                # Decorate the `initialise` method to log
                # exceptions and fail gracefully.
                cls.initialise = cls.exception_handler(cls.initialise)  # type: ignore[attr-defined] # reason: decorate all child run methods # noqa: B950
            else:
                raise AttributeError(
                    f"Worker algorithm {cls.__name__} does not have a `initialise`"
                    f" method; all subclasses should implement a `initialise()`."
                )

    def _apply_pod_dp(self, pod_dp: Optional[DPPodConfig]) -> None:
        """Applies pod-level Differential Privacy constraints.

        Subclasses should override this method if DP is supported.

        Args:
            pod_dp: The pod DP constraints to apply or None if no constraints.
        """
        pass


# The mutable underlying dict that holds the registry information
_registry: Dict[str, Type[_BaseAlgorithmFactory]] = {}
# The read-only version of the registry that is allowed to be imported
registry: Mapping[str, Type[_BaseAlgorithmFactory]] = MappingProxyType(_registry)


class _BaseAlgorithmFactory(ABC, _RolesMixIn, _BaseSerializableObjectMixIn):
    """Base algorithm factory from which all other algorithms must inherit.

    Attributes:
       class_name: The name of the algorithm class.
    """

    fields_dict: ClassVar[T_FIELDS_DICT] = {}
    nested_fields: ClassVar[T_NESTED_FIELDS] = {}

    def __init__(self, **kwargs: Any):
        self.class_name = AlgorithmType[type(self).__name__].value
        super().__init__(**kwargs)

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        if not inspect.isabstract(cls):
            logger.debug(f"Adding {cls.__name__}: {cls} to Algorithm registry")
            _registry[cls.__name__] = cls



"""Types related to identity verification flow."""
from __future__ import annotations

from asyncio import Task
from typing import Protocol, runtime_checkable

from bitfount.federated.transport.modeller_transport import _ModellerMailbox


class _ResponseHandler(Protocol):
    """Protocol describing a verification response handler."""

    async def handle(self, modeller_mailbox: _ModellerMailbox) -> None:
        """Handles identity verification response messages."""
        ...


@runtime_checkable
class _HasWebServer(Protocol):
    """Protocol describing identity verification methods that need a web server."""

    def start_server(self) -> Task:
        """Starts the identity verification process web server.

        To avoid blocking the main thread, this should be run as a background
        asyncio.Task.

        This is particularly important for identity verification methods that require
        a webserver connection as part of the _flow_, as this can take some time to
        start up.
        """
        ...

    async def stop_server(self) -> None:
        """Stops the identity verification process web server.

        It must be possible to call this method multiple times, simply doing nothing
        on subsequent calls.
        """
        ...



"""Base models and helper classes using PyTorch as the backend."""
from __future__ import annotations

from abc import abstractmethod
import contextlib
from functools import partial
import inspect
from itertools import chain
import logging
import os
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import desert
from marshmallow import fields
import numpy as np
from opacus import GradSampleModule, PrivacyEngine
from opacus.accountants import RDPAccountant
from opacus.accountants.utils import get_noise_multiplier
from opacus.validators import ModuleValidator, register_module_fixer
from opacus.validators.errors import UnsupportedModuleError
import pandas as pd
import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.callbacks.progress import TQDMProgressBar
from pytorch_lightning.loggers import LightningLoggerBase, TensorBoardLogger
from sklearn.base import BaseEstimator
import torch
from torch import Tensor
from torch import nn as nn

# Convention is to import functional as F
# noinspection PyPep8Naming
from torch.nn import functional as F
import torch.optim as optimizers
import torch.optim.lr_scheduler as schedulers
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader
import torch_optimizer as torch_optimizers

from bitfount.backends.pytorch.data.dataloaders import (
    PyTorchBitfountDataLoader,
    PyTorchIterableBitfountDataLoader,
    _BasePyTorchBitfountDataLoader,
)
from bitfount.backends.pytorch.federated.mixins import _PyTorchDistributedModelMixIn
from bitfount.backends.pytorch.types import (
    ImgAndTabBatch,
    ImgAndTabDataSplit,
    ImgDataReturnType,
    ImgFwdTypes,
    ImgXorTabDataSplit,
    TabDataReturnType,
    TabFwdTypes,
    TabxorImgBatch,
    _AdaptorForPyTorchTensor,
)
from bitfount.backends.pytorch.utils import LoggerType, _autodetect_gpu
from bitfount.backends.pytorch.weight_clipper import _PytorchParamConstraint
from bitfount.config import BITFOUNT_LOGS_DIR
from bitfount.data.databunch import BitfountDataBunch
from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasets import _BitfountDataset
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.types import DataSplit, _SingleOrMulti
from bitfount.federated.privacy.differential import (
    DPModellerConfig,
    _DifferentiallyPrivate,
)
from bitfount.metrics import Metric, MetricCollection
from bitfount.models.base_models import (
    ClassifierMixIn,
    LoggerConfig,
    ModelContext,
    NeuralNetworkMixIn,
    NeuralNetworkPredefinedModel,
    Optimizer,
    Scheduler,
    _BaseModel,
)
from bitfount.types import T_DTYPE, T_FIELDS_DICT, _StrAnyDict, _TensorLike
from bitfount.utils import _merge_list_of_dicts, delegates, seed_all

logger = logging.getLogger(__name__)

_OptimizerType = Union[torch_optimizers.Optimizer, optimizers.Optimizer]

_OPTIMIZERS: Dict[str, Type[_OptimizerType]] = {
    name: class_
    # If there are name clashes, preference is given to the original torch version
    for name, class_ in dict(
        cast(
            Iterator[Tuple[str, Any]],
            chain.from_iterable(
                d.items() for d in (vars(torch_optimizers), vars(optimizers))
            ),
        )
    ).items()
    if inspect.isclass(class_)
    and issubclass(class_, (optimizers.Optimizer, torch_optimizers.Optimizer))
    and not inspect.isabstract(class_)
    and name != "Optimizer"
}

_SCHEDULERS: Dict[str, Type[_LRScheduler]] = {
    name: class_
    for name, class_ in vars(schedulers).items()
    if inspect.isclass(class_)
    and issubclass(class_, _LRScheduler)
    and not inspect.isabstract(class_)
    and name != "_LRScheduler"
}

_TORCH_DTYPES: Dict[int, torch.dtype] = {
    16: torch.float16,
    32: torch.float32,
    64: torch.float64,
}
_STEP_OUTPUT = Union[torch.Tensor, _StrAnyDict]  # From pl.LightningModule


def _calculate_embedding_sizes(
    categorical_feature_sizes: Iterable[int],
) -> List[Tuple[int, int]]:
    """Calculate embedding sizes.

    The formula for determining the size of the embeddings is determined by
    empirical evidence alone and borrowed from the fast.ai library.

    Args:
        categorical_feature_sizes: An iterable of the number of categories in each
            categorical feature.

    Returns:
        A list of (number of categories, embedding vector length) pairs for each
        categorical feature.
    """
    embedding_sizes = [
        (n_categories, min(600, round(1.6 * n_categories**0.56)))
        for n_categories in categorical_feature_sizes
    ]
    return embedding_sizes


# Module fixer implementation that is more friendly with our expected layer sizes, etc.
# We register this as the preferred module replacer for BatchNorm layers in Opacus.
@register_module_fixer(  # type: ignore[misc] # Reason: untyped decorator but not ours
    [nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm]
)
def _fix_batch_norm(
    module: Union[nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm]
) -> nn.GroupNorm:
    # A default value of 32 is chosen for the number of groups based on the paper
    # *Group Normalization* https://arxiv.org/abs/1803.08494.
    # If this is not appropriate for the num_features in the layer we then try to
    # find an alternative.
    num_groups = min(32, module.num_features)
    while num_groups > 1:
        # If num_features is divisible by num_groups we can use it
        if module.num_features % num_groups == 0:
            break
        # Otherwise, try the next number down
        num_groups -= 1
    else:
        # If we reach the end of the while-loop we couldn't find a compatible
        # number of groups.
        raise UnsupportedModuleError(
            "Unable to find compatible number of groups for GroupNorm replacement "
            "of BatchNorm. Tried [2,32]."
        )

    return nn.GroupNorm(num_groups, module.num_features, affine=module.affine)


@delegates()
class PyTorchClassifierMixIn(ClassifierMixIn):
    """MixIn for PyTorch classification problems.

    PyTorch classification models must have this class in their inheritance hierarchy.
    """

    def _do_output_activation(self, output: torch.Tensor) -> torch.Tensor:
        """Perform final activation function on output."""
        if self.multilabel:
            return torch.sigmoid(output)
        else:
            return F.softmax(output, dim=1)


@delegates()
class _PyTorchNeuralNetworkMixIn(NeuralNetworkMixIn):
    """All Pytorch Neural Networks must inherit from this abstract class.

    Specifies model structure and hyperparameters.
    """

    @staticmethod
    def _get_optimizer(optimizer: Optimizer) -> Callable[..., _OptimizerType]:
        """Returns appropriate optimizer class."""
        if optimizer.name in _OPTIMIZERS:
            return partial(_OPTIMIZERS[optimizer.name], **optimizer.params)

        raise ValueError(
            "Optimizer not supported.",
            "Please provide one supported by 'torch' or 'torch_optimizer'.",
        )

    @staticmethod
    def _get_scheduler(scheduler: Scheduler) -> Callable[..., _LRScheduler]:
        """Returns appropriate scheduler class."""
        if scheduler.name in _SCHEDULERS:
            return partial(_SCHEDULERS[scheduler.name], **scheduler.params)

        raise ValueError(
            "Scheduler not supported.",
            "Please provide one supported by 'torch'.",
        )

    @staticmethod
    def _get_early_stopping_callback(
        early_stopping_params: Optional[_StrAnyDict] = None,
    ) -> EarlyStopping:
        """Returns EarlyStopping Callback."""
        if early_stopping_params is None:
            early_stopping_params = dict(
                monitor="validation_loss",
                min_delta=0.00,
                patience=2,
                verbose=True,
                mode="min",
            )

        return EarlyStopping(**early_stopping_params)


T_PYTORCH = TypeVar("T_PYTORCH", bound="BasePyTorchModel")


@delegates()
class BasePyTorchModel(
    _PyTorchDistributedModelMixIn,
    _PyTorchNeuralNetworkMixIn,
    _DifferentiallyPrivate,
    _BaseModel,
    Generic[T_PYTORCH],
    pl.LightningModule,
):
    """Implements a Neural Network in PyTorch Lightning.

    Args:
        model_name: Used for tensorboard logging. Model name will be left blank and
            default to "default" if none provided. Optional.
        model_version: Used for tensorboard logging.If version is not specified the
            logger inspects the save directory for existing versions, then automatically
            assigns the next available version. If it is a string then it is used as the
            run-specific subdirectory name, otherwise 'version_${version}' is used.
            Optional.
        stochastic_weight_avg: Whether to use [Stochastic Weight Averaging (SWA)](
            https://pytorch.org/blog/pytorch-1.6-now-includes-stochastic-\
weight-averaging).
        logger_config: Logger configuration. Optional. Will default to Tensorboard if
            not provided.

    Attributes:
        epochs: Number of epochs to train for when calling fit.
        model_name: Name of the model.
        model_version: Version of the model.
        steps: Number of steps to train for when calling fit.
        stochastic_weight_avg: Whether to use Stochastic Weight Averaging (SWA).
    """

    # Validate attributes
    _validation_results: List[Dict[str, str]]

    # Test attributes
    _test_preds: Optional[List[np.ndarray]]
    _test_targets: Optional[List[np.ndarray]]

    train_dl: _BasePyTorchBitfountDataLoader
    # The below fields are left as nested due the
    # `load_only`/`dump_only` methods they have
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "_dp_config": fields.Nested(
            desert.schema(DPModellerConfig),
            allow_none=True,
            data_key="dp_config",
            dump_only=True,
        ),
        "dp_config": fields.Nested(
            desert.schema(DPModellerConfig),
            allow_none=True,
            data_key="dp_config",
            load_only=True,
        ),
    }

    def __init__(
        self,
        model_name: Optional[str] = None,
        model_version: Optional[Union[int, str]] = None,
        stochastic_weight_avg: bool = False,
        logger_config: Optional[LoggerConfig] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        self.model_name = model_name
        self.model_version = model_version
        self.stochastic_weight_avg = stochastic_weight_avg
        # Set training attributes
        self._model: Optional[nn.Module] = None  # actual underlying model
        self._pl_logger = self._get_logger_from_config(logger_config)
        self._pl_trainer: pl.Trainer = self.trainer_init()  # cannot be "self.trainer"
        self._trained_on_previous_batch: bool = False
        self._total_num_batches_trained: int = 0
        # Set placeholders for DP
        self._dp_engine: Optional[PrivacyEngine] = None
        # TODO: [BIT-1614] Re-enable user-provided noise_multiplier after the
        #       post-update privacy violation issue is resolved.
        self._noise_multiplier: Optional[float] = None
        # Store initial values for steps/epochs as these may get overwritten in
        # a set_model_training_iterations() call.
        # TODO: [BIT-1614] Re-enable user-provided noise_multiplier after the
        #       post-update privacy violation issue is resolved. These are only
        #       needed for the alternative approach of get_noise_multiplier().
        self._total_epochs = self.epochs
        self._total_steps = self.steps

    @abstractmethod
    def _create_model(self) -> nn.Module:
        """Creates and returns model."""
        raise NotImplementedError

    @abstractmethod
    def _split_dataloader_output(
        self,
        data: Union[
            ImgAndTabDataSplit,
            ImgXorTabDataSplit,
        ],
    ) -> Union[ImgDataReturnType, TabDataReturnType]:
        """Splits the data from the dataloader into X, y."""
        raise NotImplementedError

    @abstractmethod
    def _do_output_activation(self, output: torch.Tensor) -> torch.Tensor:
        """Perform final activation function on output.

        Should be provided by MixIn class.
        """
        raise NotImplementedError

    def tensor_precision(self) -> T_DTYPE:
        """Returns tensor dtype used by Pytorch Lightning Trainer.

        :::note

        Currently only 32-bit training is supported.

        :::

        Returns:
            Pytorch tensor dtype.
        """
        # TODO: [BIT-727] support non-32 bit training
        return cast(T_DTYPE, _TORCH_DTYPES[cast(int, self._pl_trainer.precision)])

    def initialise_model(
        self,
        data: Optional[BaseSource] = None,
        context: Optional[ModelContext] = None,
    ) -> None:
        """Any initialisation of models/dataloaders to be done here.

        Initialises the dataloaders and sets `self._model` to be the output from
        `self.create_model`. Any initialisation ahead of federated or local training,
        serialization or deserialization should be done here.

        Args:
            data: The datasource for model training. Defaults to None.
            context: Indicates if the model is running as a modeller or worker.
                If None, there is no difference between modeller and worker.
        """
        self._initialised = True
        seed_all(self.seed)
        # Set initialisation context
        self._context = context
        if self._context == ModelContext.MODELLER:
            # In a distributed setting, the Modeller needs to first initialise its
            # own model before it can be used. The pod identifier needs to be set
            # before the model is initialised so the the relevant details can be
            # retrieved from the schema. For this we just use the first pod
            # identifier specified in the datastructure as it is assumed the the
            # schemas for all the Pods are the same.
            pod_identifiers = self.datastructure.get_pod_identifiers()
            if pod_identifiers:
                self.set_pod_identifier(pod_identifiers[0])

        if data is not None:
            if self.datastructure.query:
                table_schema = self.datastructure._override_schema(
                    pod_identifier=self.pod_identifier
                )
                self.databunch = BitfountDataBunch(
                    data_structure=self.datastructure,
                    schema=table_schema,
                    datasource=data,
                )
            elif self.datastructure.table:
                if context:
                    table_schema = self.schema.get_table_schema(
                        self.datastructure.get_table_name(self.pod_identifier)
                    )
                    self.databunch = BitfountDataBunch(
                        data_structure=self.datastructure,
                        schema=table_schema,
                        datasource=data,
                    )
                else:
                    # For local training, we can add the datasource to the schema.
                    data.load_data(table_name=self.datastructure.table)
                    self._add_datasource_to_schema(data)

            # Initialise model details
            if self._context != ModelContext.MODELLER:
                self._set_dataloaders(self.batch_size)
        self._opt_func: Callable[..., _OptimizerType] = self._get_optimizer(
            self.optimizer
        )
        self._scheduler_func: Optional[Callable[..., _LRScheduler]] = None
        if isinstance(self.scheduler, Scheduler):
            self._scheduler_func = self._get_scheduler(self.scheduler)

        self._model = self._create_model()

        # Initialise DP if requested
        self._initialise_differential_privacy()

    def _get_logger_from_config(
        self, logger_config: Optional[LoggerConfig]
    ) -> LightningLoggerBase:
        """Initialises the logger for the model."""
        if logger_config is None:
            ml_logger = TensorBoardLogger(
                name=self.model_name,
                version=self.model_version,
                save_dir=str(BITFOUNT_LOGS_DIR),
            )
        else:
            if logger_config.name == "Neptune":
                ml_logger = LoggerType[logger_config.name].value(**logger_config.params)
            else:
                if logger_config.save_dir is None:
                    logger_config.save_dir = BITFOUNT_LOGS_DIR
                ml_logger = LoggerType[logger_config.name].value(
                    save_dir=logger_config.save_dir, **logger_config.params
                )
        return ml_logger

    def _initialise_differential_privacy(self) -> None:
        """Initialises the Differential Privacy Engine if requested."""
        conf = self._dp_config
        if conf:
            # Check DP can be used with model
            if not self._model:
                raise ValueError("Model uninitialized")
            validation_errors = ModuleValidator.validate(self._model)
            if validation_errors:
                # If autofix, try to do so
                if conf.auto_fix:
                    errors_str = "\n\t".join(repr(err) for err in validation_errors)
                    # Logging the opacus errors for debugging purposes in our logs.
                    logger.debug(
                        f"Incompatible modules detected in model: \n\t{errors_str}"
                    )
                    # Still raise a warning.
                    logger.warning(
                        "Some of the modules used in the model are incompatible "
                        "with `opacus`, attempting to autofix."
                    )
                    self._model = ModuleValidator.fix_and_validate(self._model)
                else:
                    raise UnsupportedModuleError(validation_errors)

            # Create engine if not on modeller
            if self._context != ModelContext.MODELLER:
                self._dp_engine = PrivacyEngine(secure_mode=True)

            # Change dataloader if not on modeller
            if self._context != ModelContext.MODELLER:
                # Recreate train dataloader with shuffled data, which is required for
                # privacy guarantees. It will be replaced with a CSPRNG shuffler later
                # by Opacus itself.
                # We use the private method to do this to ensure the created dataloader
                # uses the expected dataset and batch size.
                logger.info("Creating shuffled training dataloader")
                self.train_dl.shuffle = True
                if isinstance(self.train_dl, PyTorchIterableBitfountDataLoader):
                    self.train_dl.secure_rng = True
                elif isinstance(self.train_dl, PyTorchBitfountDataLoader):
                    self.train_dl.dataloader = self.train_dl.get_pytorch_dataloader()

            # In order for the model to be compatible on both modeller and worker
            # we need to manually make it privacy compatible by wrapping it in a
            # GradSampleModule.
            # This is the same approach taken in PrivacyEngine.make_private() and
            # doing it now, ahead of time, is not incompatible with that call as it's
            # able to detect that the model is already wrapped.
            # We use batch_first=True as that's the default value in
            # PrivacyEngine.make_private() and we don't expose the ability to
            # specify a different value. make_private() will also raise an error
            # if the values of batch_first or loss_reduction differ which makes
            # this forward-compatible with any changes to the default values in
            # make_private().
            if not self._model:
                raise ValueError("Model uninitialized")
            self._model = GradSampleModule(
                self._model,
                batch_first=True,
                loss_reduction=conf.loss_reduction,
            )

            # Create noise multiplier (only if on worker, as needs dataloader info)
            if self._context != ModelContext.MODELLER:
                # TODO: [BIT-1614] Re-enable user-provided noise_multiplier after the
                #       post-update privacy violation issue is resolved.
                target_epsilon = conf.epsilon
                target_delta = conf.delta
                epochs = self._total_epochs
                steps = self._total_steps

                self._noise_multiplier = get_noise_multiplier(
                    target_epsilon=target_epsilon,
                    target_delta=target_delta,
                    # This is equivalent to batch_size/total_num_samples as
                    # len(self.train_dl) already returns the number of batches.
                    sample_rate=1 / len(self.train_dl),
                    epochs=epochs,
                    steps=steps,
                )

                logger.info(
                    f"Using DP noise multiplier of {self._noise_multiplier} "
                    f"({target_epsilon=}, {target_delta=}, {epochs=}, {steps=})"
                )

    def _privacy_guarantees_exceeded(self) -> bool:
        """Checks whether the privacy spent exceeds the allowed max.

        Returns: True if maximum privacy budget exceeded, False if not or if DP not
            being checked.
        """
        # If monitoring DP, perform checks
        if self._dp_engine:
            if not self._dp_config:
                raise ValueError(
                    "DP Engine created but no configuration could be found."
                )

            epsilon = self._dp_engine.get_epsilon(self._dp_config.delta)

            # Check privacy constraints
            if epsilon >= self._dp_config.epsilon:
                return True

        # If not exceeded or not being checked
        return False

    def trainer_init(self) -> pl.Trainer:
        """Initialises the Lightning Trainer for this model.

        Documentation for pytorch-lightning trainer can be found here:
        https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html

        Returns:
            The pytorch lightning trainer.
        """
        callbacks: List[Callback] = [TQDMProgressBar(refresh_rate=1)]

        if self.early_stopping:
            callbacks.append(
                self._get_early_stopping_callback(self.early_stopping.params)
            )

        # TODO: [BIT-1412] allow the user to provide these arguments in the constructor
        trainer = pl.Trainer(
            max_epochs=self.epochs or -1,
            max_steps=self.steps or -1,
            deterministic=True,
            stochastic_weight_avg=self.stochastic_weight_avg,
            auto_lr_find=True,
            logger=self._pl_logger,
            callbacks=callbacks,
            gpus=_autodetect_gpu(),  # will be 1 (first GPU) or 0 (CPU)
            check_val_every_n_epoch=1,
        )
        return trainer

    def train_dataloader(self) -> Optional[BitfountDataLoader]:  # type: ignore[override] # Reason: see below  # noqa: B950
        """Returns training dataloader."""
        # We override the dataloader return annotation as the LightningModule
        # expects a pytorch DataLoader, and we return PyTorchBitfountDataLoader
        return self.train_dl

    def val_dataloader(self) -> Optional[BitfountDataLoader]:  # type: ignore[override] # Reason: see below  # noqa: B950
        """Returns validation dataloader."""
        # We override the dataloader return annotation as the LightningModule
        # expects a pytorch DataLoader, and we returnPyTorchBitfountDataLoader
        return self.validation_dl

    def test_dataloader(self) -> Optional[BitfountDataLoader]:  # type: ignore[override] # Reason: see below  # noqa: B950
        """Returns test dataloader."""
        # We override the dataloader return annotation as the LightningModule
        # expects a pytorch DataLoader, and we return PyTorchBitfountDataLoader
        return self.test_dl

    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Serialize model to file with provided `filename`.

        Args:
            filename: Path to file to save serialized model.
        """
        if not self._initialised:
            logger.info("Model not yet initialised. Auto-initialising model.")
            self.initialise_model()
        # Model has been initialised, assuring mypy of this
        assert self._model is not None  # nosec[assert_used]
        torch.save(self._model.state_dict(), filename)

    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Deserialize model.

        :::danger

        This should not be used on a model file that has been received across a
        trust boundary due to underlying use of `pickle` by `torch`.

        :::

        Args:
            filename: Path to file containing serialized model.
        """
        if not self._initialised:
            logger.info("Model not yet initialised. Auto-initialising model.")
            self.initialise_model()
        # Model has been initialised, assuring mypy of this
        assert self._model is not None  # nosec[assert_used]
        self._model.load_state_dict(torch.load(filename))

    @staticmethod
    def _get_layer_sizes(
        layers: List[int], input_size: int
    ) -> Tuple[Tuple[int, int], ...]:
        """Converts layer sizes to in a tuple of # in and out parameters."""
        layer_sizes = []
        for idx, size in enumerate(layers):
            if idx == 0:
                layer_sizes.append((input_size, size))
            else:
                layer_sizes.append((layers[idx - 1], size))

        return tuple(layer_sizes)

    def _get_loss(
        self,
        output: torch.Tensor,
        target: torch.Tensor,
        loss_modifiers: Tuple[torch.Tensor, ...],
    ) -> Tensor:
        """Computes the appropriate (weighted) loss.

        Args:
            output: The model output.
            target: The expected output.
            loss_modifiers: A tuple of tensors representing additional loss modifiers:
                - 0: weightings for each sample in this batch
                - 1: categories

        Returns:
            A scalar tensor representing the loss.
        """
        sample_weight: torch.Tensor = loss_modifiers[0]
        categories: torch.Tensor = loss_modifiers[1]
        criterion = cast(Callable, self.loss_func)(reduction="none")
        # Handle categorical classification
        output = self._handle_categories_or_tuple(output, categories)

        if self.loss_func == nn.CrossEntropyLoss:
            loss: torch.Tensor = criterion(output, target.long())
        elif self.loss_func == nn.BCEWithLogitsLoss:
            loss = criterion(output, target.type_as(output))
            loss = loss.mean(dim=1)
        else:
            raise ValueError(
                f"Loss function {type(self.loss_func).__name__} is unsupported."
            )

        # Calculate weighted mean loss
        loss = loss * sample_weight.squeeze() / sample_weight.squeeze().sum()
        loss = loss.sum()

        return loss

    @staticmethod
    def _handle_categories_or_tuple(
        output: Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]],
        categories: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Handles conversion of output if categorical or tensor."""
        if categories is not None:
            # get output corresponding to right head depending on label
            categories = categories.unsqueeze(dim=1)
            output_to_stack: Tuple[Tensor, ...]
            if not isinstance(output, tuple):
                output_to_stack = (output,)
            else:
                output_to_stack = output
            stacked_output = torch.stack(output_to_stack, 1)
            categories = categories.expand_as(stacked_output)
            output = stacked_output.gather(1, categories)[:, 1]
        # Handle tuple output
        elif isinstance(output, tuple):
            output = output[0]
        return output

    def _compute_training_metrics(
        self,
        outputs: np.ndarray,
        targets: np.ndarray,
        training_metrics: Optional[MutableMapping[str, Metric]] = None,
    ) -> Dict[str, str]:
        """Computes metrics to be printed whilst training."""
        m = MetricCollection.create_from_model(self, training_metrics)
        results = m.compute(targets, outputs, threshold=0.5)
        self.log_dict(results, on_epoch=True, logger=True)
        results_with_str_values: Dict[str, str] = {}
        for k, v in results.items():
            results_with_str_values[k] = str(v)
        return results_with_str_values

    def configure_optimizers(
        self,
    ) -> Union[_OptimizerType, Tuple[List[_OptimizerType], List[_LRScheduler]]]:
        """Configures the optimizer(s) and scheduler(s) for backpropagation.

        Returns:
            Either the optimizer of your choice or a tuple of optimizers and learning
            rate schedulers.
        """
        parameters = filter(lambda p: p.requires_grad, self.parameters())
        optimizer: _OptimizerType = self._opt_func(parameters)

        # Attach PrivacyEngine if present
        if self._dp_engine:
            if not self._dp_config:
                raise ValueError(
                    "DP Engine created but no configuration could be found."
                )

            if isinstance(self.train_dl, PyTorchBitfountDataLoader):
                dataloader = self.train_dl.dataloader
            elif isinstance(self.train_dl, PyTorchIterableBitfountDataLoader):
                # TODO: [BIT-1685] We are using a custom dataloader which is not what
                # opacus expects so we need to have a workaround for this `make_private`
                # method call. We likely do not even need to replace the dataloader
                # anyway so if possible we should make the model and optimizer private
                # individually.
                dataloader = self.train_dl  # type: ignore[assignment] # Reason: See above # noqa: B950

            # Create DP-sensitive (model, optimizer, dataloader)
            if not self._model:
                raise ValueError("Model uninitialized")
            dp_model, dp_optimizer, dp_dataloader = self._dp_engine.make_private(
                module=self._model,
                optimizer=optimizer,
                data_loader=dataloader,
                # If self._dp_engine exists then self._noise_multiplier will
                # also have been set, and this method is only used on the worker.
                noise_multiplier=cast(float, self._noise_multiplier),
                max_grad_norm=self._dp_config.max_grad_norm,
                loss_reduction=self._dp_config.loss_reduction,
                # TODO: [BIT-1474] Identify what changes are needed to enable
                #       poisson_sampling=True.
                poisson_sampling=False,
            )

            logger.info("Replacing model with DP-sensitive version.")
            self._model = dp_model

            # TODO: [BIT-1685] Confirm that we don't need to replace the dataloader if
            # it is iterable
            if isinstance(self.train_dl, PyTorchBitfountDataLoader):
                logger.info("Replacing dataloader with DP-sensitive version.")
                self.train_dl.dataloader = dp_dataloader

            optimizer = dp_optimizer
        if self._scheduler_func:
            scheduler = self._scheduler_func(optimizer)
            return [optimizer], [scheduler]

        return optimizer

    def forward(self, x: Union[ImgFwdTypes, TabFwdTypes]) -> Any:  # type: ignore[override] # reason: see below # noqa: B950
        """Performs a forward pass of the underlying model.

        Args:
            x: Input to the model.

        Returns:
            Output of the model.

        """
        # Signature incompatible with pl.LightningModule:
        # Superclass: def forward(self, *args: Any, **kwargs: Any) -> Any
        # Subclass: def forward(self, x: Union[ImgFwdTypes, TabFwdTypes]) -> Any
        return self._model(x)  # type: ignore[misc] # reason: the model is set by this stage # noqa: B950

    def _skip_training_batch(self, batch_idx: int) -> bool:
        """Checks if the current batch from the training set should be skipped.

        This is a workaround for the fact that PyTorch Lightning starts the Dataloader
        iteration from the beginning every time `fit` is called. This means that if we
        are training in steps, we are always training on the same batches. So this
        method needs to be called at the beginning of every `training_step` to skip
        to the right batch index.

        Args:
            batch_idx: the index of the batch from `training_step`.

        Returns:
            True if the batch should be skipped, otherwise False.
        """
        # TODO: [BIT-1237] remove this code block and find a better way to do this that
        # doesn't involve loading every batch into memory until we get to the right one
        if self.steps:
            # If we have trained on the previous batch, we can avoid the checks because
            # it means we have already reached the target start batch.
            if not self._trained_on_previous_batch:
                if (self.steps != self._pl_trainer.max_steps) and (
                    batch_idx < (self._total_num_batches_trained % len(self.train_dl))
                ):
                    return True
                else:
                    self._trained_on_previous_batch = True

            # `_total_num_batches_trained` hasn't been incremented yet so we need to add
            # 1 here to get the correct batch number.
            if self._total_num_batches_trained + 1 == self._pl_trainer.max_steps:
                self._trained_on_previous_batch = False

        if not self._pl_trainer.sanity_checking:
            self._total_num_batches_trained += 1

        return False

    def training_step(self, batch: Tuple[Union[TabxorImgBatch, ImgAndTabBatch], torch.Tensor], batch_idx: int) -> Optional[torch.Tensor]:  # type: ignore[override] # Reason: see below # noqa: B950
        """Training step.

        Args:
            batch: The batch to be trained on.
            batch_idx: The index of the batch to be trained on from the train
                dataloader.

        Returns:
            The loss from this batch as a `torch.Tensor`.
        """
        # Override the pl.lightning method, as it requires *args
        # and **kwargs as arguments whereas we will always have
        # the batch and the batch_idx as args.
        # `data` contains the X batch and often some additional information.
        # This can be split into relevant chunks by the `split_dataloader_output()`
        # method, which details how to perform that split.
        if self.param_clipping:
            # Make sure that the parameters stay within required bounds
            # for SecureAggregation
            clipper = _PytorchParamConstraint(**cast(dict, self.param_clipping))
            # If we are in this check, self.param_clipping is
            # a dictionary so it is safe to cast

            # Check for BatchNorm modules
            modules_to_clip = [
                mod_name
                for mod_name, modules in self._model._modules.items()  # type: ignore[union-attr] # Reason: _model is initialized before training  # noqa: B950
                if "BatchNorm" in str(modules)
            ]
            # Apply the parameter contraint to all batch norm modules
            for mod in modules_to_clip:
                self._model._modules[mod].apply(clipper)  # type: ignore[union-attr] # Reason: _model is initialized before training  # noqa: B950

        if self._skip_training_batch(batch_idx):
            # Returning `None` to skip this batch
            return None

        data, y = batch
        x, *loss_modifiers = self._split_dataloader_output(data)
        # Perform forward pass and get loss
        y_hat = self(x)
        loss = self._get_loss(
            output=y_hat,
            target=y,
            loss_modifiers=cast(Tuple[torch.Tensor, ...], loss_modifiers),
        )
        self.log(
            "train_loss", loss, on_step=True, on_epoch=True, prog_bar=True, logger=True
        )
        return loss

    def training_epoch_end(self, kwargs: Any) -> None:
        """Extract gradient from weights after training."""
        if self.param_clipping:
            # Make sure that the parameters stay within required bounds
            # for SecureAggregation
            clipper = _PytorchParamConstraint(**cast(dict, self.param_clipping))
            # If we are in this check, self.param_clipping
            # is a dictionary so it is safe to cast

            # Check for BatchNorm modules
            modules_to_clip = [
                mod_name
                for mod_name, modules in self._model._modules.items()  # type: ignore[union-attr] # Reason: _model is initialized before training  # noqa: B950
                if "BatchNorm" in str(modules)
            ]
            # Apply the parameter contraint to all batch norm modules
            for mod in modules_to_clip:
                self._model._modules[mod].apply(  # type: ignore[union-attr] # Reason: _model is initialized before training  # noqa: B950
                    clipper
                )

        try:
            # For now, this will fail for the resnet model since it is defined with
            # no layers, so wrapping in a try...except.
            self._training_epoch_end_gradients = [
                layer.weight.grad.detach().clone()
                for layer in self._model.layers  # type: ignore[union-attr] # Reason: _model is initialized before training  # noqa: B950
            ]
        except AttributeError:
            pass

    def on_train_batch_start(  # type: ignore[override] # reason: incompatible with supertype return type # noqa: B950
        self,
        batch: _SingleOrMulti[_SingleOrMulti[np.ndarray]],
        batch_idx: int,
    ) -> Optional[Literal[-1]]:
        """Checks if any privacy guarantees have been exceeded and stops training if so.

        Args:
            batch: The batch to be trained on.
            batch_idx: The index of the batch to be trained on from the train
                dataloader.
            dataloader_idx: The index of the dataloader from which the batch was
                taken. This is useful for multi-dataloader training.

        Returns:
            -1 if the entire epoch should be skipped, otherwise None.
        """
        # Check privacy constraints and set stopping criterion as needed
        if self._privacy_guarantees_exceeded():
            self._pl_trainer.should_stop = True  # stops later epochs from running
            return -1  # this skips training for the rest of the current epoch
        else:
            return None

    def validation_step(self, batch: Tuple[Union[TabxorImgBatch, ImgAndTabBatch], torch.Tensor], batch_idx: int) -> Dict[str, Tensor]:  # type: ignore[override] # Reason: see below # noqa: B950
        """Validation step.

        Args:
            batch: The batch to be evaluated.
            batch_idx: The index of the batch to be evaluated from the validation
                dataloader.

        Returns:
            A dictionary of strings and values that should be averaged at the end of
            every epoch and logged e.g. `{"validation_loss": loss}`. These will be
            passed to the `validation_epoch_end` method.
        """
        # Override the pl.lightning method, as it requires
        # *args and **kwargs as arguments whereas we will
        # always have the batch and the batch_idx as args
        # Extract X, y and other data from batch
        data, y = batch
        x, *loss_modifiers = self._split_dataloader_output(data)
        # Get validation output and loss
        y_hat = self(x)
        loss = self._get_loss(
            output=y_hat,
            target=y,
            loss_modifiers=cast(Tuple[torch.Tensor, ...], loss_modifiers),
        )
        # Handle categorical data
        categories = loss_modifiers[-1]
        y_hat = self._handle_categories_or_tuple(y_hat, categories)

        y_hat = self._do_output_activation(y_hat)

        # Log and return desired outputs
        self.log("validation_loss", loss, on_epoch=True, prog_bar=True, logger=True)
        return {"validation_loss": loss, "outputs": y_hat, "targets": y}

    def validation_epoch_end(  # type: ignore[override] # Reason: see below
        self, outputs: List[Dict[str, Tensor]]
    ) -> None:
        """Called at the end of the validation epoch with all validation step outputs.

        Ensures that the average metrics from a validation epoch is stored. Logs results
        and also appends to `self.val_stats`.

        Args:
            outputs: List of outputs from each validation step.
        """
        # Override the pl.lightning method, as it requires a different type for outputs.
        # Merge outputs into singular lists rather than a list of dicts
        merged_outputs: Dict[str, List[torch.Tensor]] = _merge_list_of_dicts(outputs)

        # Calculate metrics from outputs. This will also log them out to self.log().
        if not self._pl_trainer.sanity_checking:
            validation_metrics = self._compute_training_metrics(
                outputs=np.asarray(
                    [i.cpu().numpy() for i in merged_outputs["outputs"]]
                ),
                targets=np.asarray(
                    [i.cpu().numpy() for i in merged_outputs["targets"]]
                ),
                training_metrics=self.metrics,
            )
            # Calculate average validation loss and add to validation metrics dictionary
            avg_val_loss = torch.mean(torch.stack(merged_outputs["validation_loss"]))
            validation_metrics["validation_loss"] = f"{avg_val_loss.item():.4f}"
            self._validation_results.append(validation_metrics)

    def test_step(self, batch: Tuple[Union[TabxorImgBatch, ImgAndTabBatch], torch.Tensor], batch_idx: int) -> _StrAnyDict:  # type: ignore[override] # Reason: see below # noqa: B950
        """Make sure to set self.preds and self.target before returning in this method.

        They will be returned by the `evaluate` method.

        Args:
            batch: The batch to be evaluated.
            batch_idx: The index of the batch to be evaluated from the test
                dataloader.

        Returns:
            A dictionary of predictions and targets. These will be passed to the
            `test_epoch_end` method.
        """
        # Override the pl.lightning method, as it requires
        # *args and **kwargs as arguments whereas we will
        # always have the batch and the batch_idx as args
        # Extract X, y and other data from batch
        data, y = batch
        x, *loss_modifiers = self._split_dataloader_output(data)

        # Get validation output and loss
        y_hat = self(x)

        # Handle categorical or tuple output
        categories = loss_modifiers[-1]
        y_hat = self._handle_categories_or_tuple(y_hat, categories)

        y_hat = self._do_output_activation(y_hat)
        # Output targets and prediction for later
        return {"predictions": y_hat, "targets": y}

    def test_epoch_end(self, outputs: List[Dict[str, torch.Tensor]]) -> None:  # type: ignore[override] # Reason: see below # noqa: B950
        """Aggregates the predictions and targets from the test set.

        Args:
            outputs: List of outputs from each test step.
        """
        # Override the pl.lightning method, as it requires a different type for outputs.
        merged_outputs: Dict[str, List[torch.Tensor]] = _merge_list_of_dicts(outputs)
        self._test_preds = [i.cpu().numpy() for i in merged_outputs["predictions"]]
        self._test_targets = [i.cpu().numpy() for i in merged_outputs["targets"]]

    def _fit_local(
        self,
        data: BaseSource,
        metrics: Optional[MutableMapping[str, Metric]] = None,
        **kwargs: Any,
    ) -> Dict[str, str]:
        """Fits the model.

        Args:
            data: The data source to be used for training the model.
            metrics: Validation metrics to print during training. Defaults to None and
                will instead use metrics appropriate to the task. Optional.

        Returns:
            Validation metrics for the final epoch
        """
        # Check model initialisation
        if self._initialised:
            logger.debug("Model already initialised, will not reinitialise.")
        else:
            logger.info("Model not yet initialised. Auto-initialising model.")
            self.initialise_model(data)

        # TODO: [BIT-499] Check DP privacy guarantees before we even start; return NULL
        #       state if exceeded.
        self._validation_results = []
        with self._use_metrics(metrics):
            # will perform self.epochs or self.steps of local training
            self._pl_trainer.fit(self)

            # If training has been specified in terms of epochs, we are guaranteed to
            # have run validation after training. However if training is specified in
            # terms of steps, we only would have run validation after training if the
            # the number of steps is the same as that in an epoch. Therefore we must
            # manually run validation to ensure we have the latest validation metrics
            # that reflect the state of the model at the end of training.
            if self.steps and self.steps % len(self.train_dl) != 0:
                self._pl_trainer.validate(self)

        try:
            # Getting the most recent set of validation metrics
            validation_results = self._validation_results[-1]
        except IndexError:
            # If we haven't had a full epoch of training, then we don't have any
            # validation results in the dictionary so we must create an empty one.
            # This is required if training is run with steps or epochs set to 0.
            validation_results = {}

        # Add DP metrics to the final epoch validation results
        if self._dp_engine:
            if not self._dp_config:
                raise ValueError(
                    "DP Engine created but no configuration could be found."
                )

            epsilon, alpha = cast(
                RDPAccountant, self._dp_engine.accountant
            ).get_privacy_spent(
                delta=self._dp_config.delta, alphas=self._dp_config.alphas
            )
            validation_results["alpha"] = str(alpha)
            validation_results["epsilon"] = str(epsilon)

        # Return the validation metrics for the final epoch
        return validation_results

    @contextlib.contextmanager
    def _use_metrics(
        self, metrics: Optional[MutableMapping[str, Metric]] = None
    ) -> Iterator[None]:
        """Context manager which temporarily sets `metrics` to `self.metrics`.

        Args:
            metrics: Validation metrics to print during training. Defaults to None and
                will instead use metrics appropriate to the task. Optional.

        Returns:
            Empty Iterator.
        """
        metrics_store: Optional[MutableMapping[str, Metric]] = None
        # Stash the old metrics if new ones provided
        if metrics:
            metrics_store = self.metrics
            self.metrics = metrics
        try:
            yield None
        finally:
            # Restore the old metrics if new ones were provided
            if metrics:
                self.metrics = metrics_store

    def _evaluate_local(
        self, test_dl: Optional[BitfountDataLoader] = None, **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        """This method runs inference on the test dataloader.

        This is done by calling `self.test_step` under the hood.
        Args:
            test_dl: Optional dataloader to run inference on which takes precedence over
                the dataloader returned by `self.test_dataloader`.

        Returns:
            A tuple of predictions and targets as numpy arrays.
        """
        # Reset metrics
        self._reset_test_attrs()

        self._pl_trainer.test(model=self, dataloaders=cast(DataLoader, test_dl))
        return np.asarray(self._test_preds), np.asarray(self._test_targets)

    def predict(self, data: BaseSource, **kwargs: Any) -> np.ndarray:
        """This method runs inference on the test data, returns predictions.

        This is done by calling `test_step` under the hood. Customise this method as you
        please but it must return a list of predictions and a list of targets. Note that
        as this is the prediction function, only the predictions are returned.

        Returns:
            A numpy array containing the prediction values.

        Raises:
            ValueError: If no data is provided to test with.
        """
        data.load_data()
        self._add_datasource_to_schema(data)
        test_dl = self.databunch.get_test_dataloader(self.batch_size)
        if isinstance(test_dl, BitfountDataLoader):
            logger.info(
                f"Using test portion of dataset for inference - this has"
                f"{len(test_dl)} records."
            )
        else:
            raise ValueError("No data provided to run inference for the model on.")

        self._pl_trainer.test(model=self, dataloaders=cast(DataLoader, test_dl))
        return cast(np.ndarray, self._test_preds)

    def _reset_test_attrs(self) -> None:
        """Resets test attributes to None."""
        self._test_preds = None
        self._test_targets = None


@delegates()
class BaseTabNetModel(
    _PyTorchDistributedModelMixIn, _PyTorchNeuralNetworkMixIn, _BaseModel
):
    """TabNet Model as described in https://arxiv.org/abs/1908.07442.

    This is a wrapper around the implementation from DreamQuark. Documentation can be
    found here: https://dreamquark-ai.github.io/tabnet.

    Args:
        virtual_batch_size: Virtual batch size used for ghost batch
            normalization.
        patience: Number of epochs before early stopping.
        scheduler: Learning rate scheduler. Defaults to None.
        scheduler_params: Learning rate scheduler params. Defaults to None.
        embedding_sizes: Embeddings sizes. Defaults to None.
        num_workers: Number of workers for torch DataLoader. Defaults to 0.
        inverse_class_weights: Inverse class weights (only for classification problems).
            Defaults to True.
        mask_type: Mask type. Defaults to "sparsemax".
        decision_prediction_layer_size: Final feedforward layer size. Defaults to 8.
        attention_embedding_layer_size: Attention embedding layer size. Defaults to 8.
        num_steps: Number of steps. Defaults to 3.


    Raises:
        ValueError: If virtual batch size > batch size.
        ValueError: If the `model_structure` does not match the TabNet model.
        ValueError: If training is specified in `steps` rather than `epochs`. Training
            steps are not supported.
    """

    train_dl: _BasePyTorchBitfountDataLoader
    fields_dict: ClassVar[T_FIELDS_DICT] = {
        "virtual_batch_size": fields.Integer(),
        "patience": fields.Integer(),
        "embedding_sizes": fields.Integer(allow_none=True),
        "num_workers": fields.Integer(),
        "inverse_class_weights": fields.Bool(),
        "mask_type": fields.String(),
        "decision_prediction_layer_size": fields.Integer(),
        "attention_embedding_layer_size": fields.Integer(),
        "num_steps": fields.Integer(),
    }

    def __init__(
        self,
        virtual_batch_size: int = 64,
        patience: int = 2,
        embedding_sizes: Optional[Union[int, List[int]]] = None,
        num_workers: int = 0,
        inverse_class_weights: bool = True,
        mask_type: str = "sparsemax",
        decision_prediction_layer_size: int = 8,
        attention_embedding_layer_size: int = 8,
        num_steps: int = 3,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        if self.steps:
            raise ValueError(
                "TabNet does not support steps. Training must be specified in epochs."
            )

        if virtual_batch_size > self.batch_size:
            raise ValueError("Virtual batch size must be smaller than batch size.")

        if (
            not isinstance(self.model_structure, NeuralNetworkPredefinedModel)
            or self.model_structure.name != "TabNet"
        ):
            raise ValueError(
                "Please provide the appropriate Model Structure.",
                "TabNet model can only create TabNet models.",
            )

        self.virtual_batch_size = virtual_batch_size
        self.patience = patience
        if not self.scheduler:
            self.scheduler = Scheduler("StepLR", {"step_size": 50, "gamma": 0.9})
        self.embedding_sizes = embedding_sizes
        self.num_workers = num_workers
        self.inverse_class_weights = inverse_class_weights
        self.mask_type = mask_type
        self.decision_prediction_layer_size = decision_prediction_layer_size
        self.attention_embedding_layer_size = attention_embedding_layer_size
        self.num_steps = num_steps

        # This is unused by the TabNet model but is required for compatibility with
        # DistributedModelProtocol
        self._total_num_batches_trained: int = 0

        # TODO: [BIT-1152] support FederatedAveraging by fixing initialisation
        logger.warning("TabNet model currently does not support FederatedAveraging.")

    @abstractmethod
    def _create_model(self) -> BaseEstimator:
        """Returns appropriate model."""
        raise NotImplementedError

    def get_param_states(self) -> Dict[str, _TensorLike]:
        """See base class."""
        aux = dict(self._model.network.state_dict())  # type: ignore[union-attr] # Reason: _model is initialised in subclass # noqa: B950

        return {key: _AdaptorForPyTorchTensor(value) for key, value in aux.items()}

    def initialise_model(
        self,
        data: Optional[BaseSource] = None,
        context: Optional[ModelContext] = None,
    ) -> None:
        """Any initialisation of models/dataloaders to be done here.

        Initialises the dataloaders and sets `self._model` to be the output from
        `self.create_model`. Any initialisation ahead of federated or local training,
        serialization or deserialization should be done here.

        Args:
            data: The datasource for model training. Defaults to None.
            context: Indicates if the model is running as a modeller or worker.
                If None, there is no difference between modeller and worker.
        """
        self._context = context
        self._initialised = True
        if self._context == ModelContext.MODELLER:
            # In a distributed setting, the Modeller needs to first initialise its
            # own model before it can be used. The pod identifier needs to be set
            # before the model is initialised so the the relevant details can be
            # retrieved from the schema. For this we just use the first pod
            # identifier specified in the datastructure as it is assumed the the
            # schemas for all the Pods are the same.
            pod_identifiers = self.datastructure.get_pod_identifiers()
            if pod_identifiers:
                self.set_pod_identifier(pod_identifiers[0])

        if data is not None:
            if self.datastructure.query:
                table_schema = self.datastructure._override_schema(
                    pod_identifier=self.pod_identifier
                )
                self.databunch = BitfountDataBunch(
                    data_structure=self.datastructure,
                    schema=table_schema,
                    datasource=data,
                )
            elif self.datastructure.table:
                if context:
                    table_schema = self.schema.get_table_schema(
                        self.datastructure.get_table_name(self.pod_identifier)
                    )
                    self.databunch = BitfountDataBunch(
                        data_structure=self.datastructure,
                        schema=table_schema,
                        datasource=data,
                    )
                else:
                    # For local training, we can add the datasource to the schema.
                    data.load_data()
                    self._add_datasource_to_schema(data)

        # TODO: [BIT-1152] initialise model without setting dataloaders
        if self._context != ModelContext.MODELLER:
            self._set_dataloaders(batch_size=self.batch_size)

        self._model = self._create_model()
        # Make sure we only use the tabular part of the data.
        train_x_df_aux = self.train_dl.get_x_dataframe()
        if isinstance(train_x_df_aux, tuple):
            train_x_df, _ = train_x_df_aux
            train_x_df = train_x_df
        else:
            train_x_df = train_x_df_aux
        # 'Train' with 0 epochs just to set model parameters
        self._model.fit(  # type: ignore[attr-defined] # reason: the model should be set by the time we fit it # noqa: B950
            train_x_df.values,
            self.train_dl.get_y_dataframe()[self.datastructure.target].values,
            max_epochs=0,
            virtual_batch_size=self.virtual_batch_size,
            num_workers=self.num_workers,
            weights=int(self.inverse_class_weights),
            drop_last=False,
        )

    def _evaluate_local(
        self, test_dl: Optional[BitfountDataLoader] = None, **kwargs: Any
    ) -> Tuple[np.ndarray, np.ndarray]:
        """This method runs inference on the test dataloader.

        This is done by calling `self.test_step` under the hood.
        Args:
            test_dl: Optional dataloader to run inference on which takes precedence over
                the dataloader returned by `self.test_dataloader`.

        Returns:
            A tuple of predictions and targets as numpy arrays.
        """
        if test_dl is None:
            if isinstance(self.test_dl, BitfountDataLoader):
                test_dl = self.test_dl
            else:
                raise ValueError("There is no test data to evaluate the model on.")
        X_aux = test_dl.get_x_dataframe()
        X: pd.DataFrame
        if isinstance(X_aux, tuple):
            X, _ = X_aux
        elif isinstance(X_aux, pd.DataFrame):
            X = X_aux
        Y = test_dl.get_y_dataframe()[self.databunch.target].values
        preds = self._model.predict_proba(  # type:ignore[union-attr] # reason: the model should be set by the time we evaluate this function # noqa: B950
            X.values
        )

        return preds, Y

    def serialize(self, filename: Union[str, os.PathLike]) -> None:
        """Serialize model to file with provided `filename`.

        Args:
            filename: Path to file to save serialized model.
        """
        if not self._initialised:
            # TODO: [BIT-1152] initialise model here instead of logging an error
            logger.error(
                "Can't serialize uninitialized model. "
                "Model can't be initialised without data."
            )
            return
        torch.save(self._model.network.state_dict(), filename)  # type: ignore[union-attr] # reason: model is initialised # noqa: B950

    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
        """Deserialize model.

        :::danger

        This should not be used on a model file that has been received across a
        trust boundary due to underlying use of `pickle` by `torch`.

        :::

        Args:
            filename: Path to file containing serialized model.
        """
        if not self._initialised:
            # TODO: [BIT-1152] initialise model here instead of logging an error
            logger.error(
                "Can't deserialize uninitialized model. "
                "Model can't be initialised without data."
            )
            return

        self._model.network.load_state_dict(torch.load(filename))  # type: ignore[union-attr] # reason: model is initialised # noqa: B950

    def _fit_local(
        self,
        data: BaseSource,
        metrics: Optional[Union[str, List[str], MutableMapping[str, Metric]]] = None,
        **kwargs: Any,
    ) -> Dict[str, str]:
        """Fit model and return results.

        Args:
            data: The data used for model training.
            metrics: list of metrics
                to use on validation set every epoch. Last metric in this list is the
                one to use for early stopping
        """
        if self._initialised:
            logger.debug("Model already initialised, will not reinitialise.")
        else:
            logger.info("Model not yet initialised. Auto-initialising model.")
            self.initialise_model(data=data)

        if metrics is None:
            metrics = ["logloss"]
        elif isinstance(metrics, str):
            metrics = [metrics]
        elif isinstance(metrics, MutableMapping):
            logger.warning(
                "TabNet models are incompatible with bitfount Metric objects;"
                " will derive TabNet Metric objects using supplied names."
            )
            metrics = [k for k in metrics.keys()]

        if "logloss" not in metrics:
            metrics.insert(0, "logloss")

        # Make sure we only use the tabular part of the data.
        train_x_df_aux = self.train_dl.get_x_dataframe()
        if isinstance(train_x_df_aux, tuple):
            train_x_df, _ = train_x_df_aux
            train_x_df = train_x_df
        else:
            train_x_df = train_x_df_aux

        if not isinstance(self.validation_dl, BitfountDataLoader):
            empty_df = pd.DataFrame(columns=train_x_df.columns)
            self.validation_dl = BitfountDataLoader(
                _BitfountDataset(
                    datasource=DataFrameSource(empty_df),
                    target=self.databunch.target,
                    selected_cols_semantic_types=self.datastructure.selected_cols_w_types,  # noqa: B950
                    selected_cols=self.datastructure.selected_cols,
                    data_split=DataSplit.VALIDATION,
                    schema=self.databunch.schema,
                )
            )
        val_x_df_aux = self.validation_dl.get_x_dataframe()
        if isinstance(val_x_df_aux, tuple):
            val_x_df, _ = val_x_df_aux
        else:
            val_x_df = val_x_df_aux

        self._model.fit(  # type: ignore[union-attr] # reason: the model should be set by the time we fit it # noqa: B950
            train_x_df.values,
            self.train_dl.get_y_dataframe()[self.databunch.target].values,
            eval_set=[
                (
                    val_x_df.values,
                    self.validation_dl.get_y_dataframe()[self.databunch.target].values,
                )
            ],
            eval_metric=metrics,
            max_epochs=self.epochs,
            patience=self.patience,
            batch_size=self.batch_size,
            virtual_batch_size=self.virtual_batch_size,
            num_workers=self.num_workers,
            weights=int(self.inverse_class_weights),
            drop_last=False,
        )

        metrics_results = {
            "validation_loss": "%.4f"
            % self._model.history["val_0_logloss"][-1]  # type: ignore[union-attr] # reason: see below # noqa: B950
            # The model should be set by the time we retrieve metrics
        }

        return metrics_results



"""Component implementations in various backends.

Modules in this package and associated subpackages must never be imported into the core
Bitfount package unless a check is performed to ensure the given backend exists.
Every subpackage in this package must have an `__init__.py` file that defines `__all__`.
"""
